"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import tensorflow.core.framework.function_pb2
import tensorflow.core.framework.graph_debug_info_pb2
import tensorflow.core.framework.node_def_pb2
import tensorflow.core.framework.versions_pb2

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class GraphDef(google.protobuf.message.Message):
    """Represents the graph of operations"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NODE_FIELD_NUMBER: builtins.int
    VERSIONS_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    LIBRARY_FIELD_NUMBER: builtins.int
    DEBUG_INFO_FIELD_NUMBER: builtins.int
    version: builtins.int
    """Deprecated single version field; use versions above instead.  Since all
    GraphDef changes before "versions" was introduced were forward
    compatible, this field is entirely ignored.
    """
    @property
    def node(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[tensorflow.core.framework.node_def_pb2.NodeDef]: ...
    @property
    def versions(self) -> tensorflow.core.framework.versions_pb2.VersionDef:
        """Compatibility versions of the graph.  See core/public/version.h for version
        history.  The GraphDef version is distinct from the TensorFlow version, and
        each release of TensorFlow will support a range of GraphDef versions.
        """

    @property
    def library(self) -> tensorflow.core.framework.function_pb2.FunctionDefLibrary:
        """ "library" provides user-defined functions.

        Naming:
          * library.function.name are in a flat namespace.
            NOTE: We may need to change it to be hierarchical to support
            different orgs. E.g.,
            { "/google/nn", { ... }},
            { "/google/vision", { ... }}
            { "/org_foo/module_bar", { ... }}
            map<string, FunctionDefLib> named_lib;
          * If node[i].op is the name of one function in "library",
            node[i] is deemed as a function call. Otherwise, node[i].op
            must be a primitive operation supported by the runtime.


        Function call semantics:

          * The callee may start execution as soon as some of its inputs
            are ready. The caller may want to use Tuple() mechanism to
            ensure all inputs are ready in the same time.

          * The consumer of return values may start executing as soon as
            the return values the consumer depends on are ready.  The
            consumer may want to use Tuple() mechanism to ensure the
            consumer does not start until all return values of the callee
            function are ready.
        """

    @property
    def debug_info(self) -> tensorflow.core.framework.graph_debug_info_pb2.GraphDebugInfo:
        """Stack traces for the nodes in this graph."""

    def __init__(
        self,
        *,
        node: collections.abc.Iterable[tensorflow.core.framework.node_def_pb2.NodeDef] | None = ...,
        versions: tensorflow.core.framework.versions_pb2.VersionDef | None = ...,
        version: builtins.int | None = ...,
        library: tensorflow.core.framework.function_pb2.FunctionDefLibrary | None = ...,
        debug_info: tensorflow.core.framework.graph_debug_info_pb2.GraphDebugInfo | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["debug_info", b"debug_info", "library", b"library", "versions", b"versions"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "debug_info", b"debug_info", "library", b"library", "node", b"node", "version", b"version", "versions", b"versions"
        ],
    ) -> None: ...

global___GraphDef = GraphDef



import re
from typing import ClassVar
from typing_extensions import deprecated

from markdown.core import Markdown

from . import util

def build_postprocessors(md: Markdown, **kwargs) -> util.Registry[Postprocessor]: ...

class Postprocessor(util.Processor):
    def run(self, text: str) -> str: ...

class RawHtmlPostprocessor(Postprocessor):
    BLOCK_LEVEL_REGEX: ClassVar[re.Pattern[str]]
    def isblocklevel(self, html: str) -> bool: ...
    def stash_to_string(self, text: str) -> str: ...

class AndSubstitutePostprocessor(Postprocessor): ...

@deprecated(
    "This class is deprecated and will be removed in the future; "
    "use [`UnescapeTreeprocessor`][markdown.treeprocessors.UnescapeTreeprocessor] instead."
)
class UnescapePostprocessor(Postprocessor):
    RE: ClassVar[re.Pattern[str]]
    def unescape(self, m: re.Match[str]) -> str: ...



from _typeshed import StrOrBytesPath

def print_as_hex(s: str) -> None: ...
def get_starting_chunk(filename: StrOrBytesPath, length: int = 1024) -> bytes: ...
def is_binary_string(bytes_to_check: bytes | bytearray) -> bool: ...



import mailbox
from _typeshed import StrPath
from collections.abc import Callable
from typing import IO, Any

DEFAULT_MAXMEM: int

class mbox_readonlydir(mailbox.mbox):
    maxmem: int
    def __init__(
        self,
        path: StrPath,
        factory: Callable[[IO[Any]], mailbox.mboxMessage] | None = None,
        create: bool = True,
        maxmem: int = 1048576,
    ) -> None: ...
    def flush(self) -> None: ...



import datetime
from typing import Final, Literal, overload
from typing_extensions import Self

from pymeeus.Angle import Angle

DAY2SEC: Final = 86400.0
DAY2MIN: Final = 1440.0
DAY2HOURS: Final = 24.0
LEAP_TABLE: Final[dict[float, int]]

class Epoch:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(
        self, year: int, month: int, day: int, /, *time: float, leap_seconds: float = 0, local: bool = False, utc: bool = False
    ) -> None: ...
    @overload
    def __init__(
        self,
        a: float | Epoch | list[float] | tuple[float, ...] | datetime.date,
        /,
        *,
        leap_seconds: float = 0,
        local: bool = False,
        utc: bool = False,
    ) -> None: ...
    def __hash__(self) -> int: ...
    @overload
    def set(self) -> None: ...
    @overload
    def set(
        self, year: int, month: int, day: int, /, *time: float, leap_seconds: float = 0, local: bool = False, utc: bool = False
    ) -> None: ...
    @overload
    def set(
        self,
        a: float | Epoch | list[float] | tuple[float, ...] | datetime.date,
        /,
        *,
        leap_seconds: float = 0,
        local: bool = False,
        utc: bool = False,
    ) -> None: ...
    @overload
    @overload
    @staticmethod
    def check_input_date(
        a: Epoch | list[float] | tuple[float, ...] | datetime.date,
        /,
        *,
        leap_seconds: float = 0,
        local: bool = False,
        utc: bool = False,
    ) -> Epoch: ...
    @overload
    @staticmethod
    def check_input_date(
        year: int, month: int, day: int, /, *, leap_seconds: float = 0, local: bool = False, utc: bool = False
    ) -> Epoch: ...
    @staticmethod
    def is_julian(year: int, month: int, day: int) -> bool: ...
    def julian(self) -> bool: ...
    @overload
    @staticmethod
    def get_month(month: float | str, as_string: Literal[True]) -> str: ...
    @overload
    @staticmethod
    def get_month(month: float | str, as_string: Literal[False] | None = False) -> int: ...
    @staticmethod
    def is_leap(year: float) -> bool: ...
    def leap(self) -> bool: ...
    @staticmethod
    def get_doy(yyyy: int, mm: int, dd: int) -> float: ...
    def doy(self) -> float: ...
    @staticmethod
    def doy2date(year: int, doy: float) -> tuple[int, int, float]: ...
    @staticmethod
    def leap_seconds(year: int, month: int) -> int: ...
    @staticmethod
    def get_last_leap_second() -> tuple[int, int, float, int]: ...
    @staticmethod
    def utc2local() -> float: ...
    @staticmethod
    def easter(year: float) -> tuple[int, int]: ...
    @staticmethod
    def jewish_pesach(year: float) -> tuple[int, int]: ...
    @staticmethod
    def moslem2gregorian(year: float, month: float, day: float) -> tuple[int, int, int]: ...
    @staticmethod
    def gregorian2moslem(year: float, month: float, day: float) -> tuple[int, int, int]: ...
    def get_date(self, *, utc: bool = False, leap_seconds: float = 0.0, local: bool = ...) -> tuple[int, int, float]: ...
    def get_full_date(
        self, *, utc: bool = False, leap_seconds: float = 0.0, local: bool = ...
    ) -> tuple[int, int, int, int, int, float]: ...
    @staticmethod
    def tt2ut(year: int, month: int) -> float: ...
    @overload
    def dow(self, as_string: Literal[True]) -> str: ...
    @overload
    def dow(self, as_string: Literal[False] | None = False) -> int: ...
    def mean_sidereal_time(self) -> float: ...
    def apparent_sidereal_time(self, true_obliquity: float | Angle, nutation_longitude: float | Angle) -> float: ...
    def mjd(self) -> float: ...
    def jde(self) -> float: ...
    def year(self) -> float: ...
    def rise_set(self, latitude: Angle, longitude: Angle, altitude: float = 0.0) -> tuple[Epoch, Epoch]: ...
    def __call__(self) -> float: ...
    def __add__(self, b: float) -> Epoch: ...
    @overload
    def __sub__(self, b: float) -> Epoch: ...
    @overload
    def __sub__(self, b: Epoch) -> float: ...
    def __iadd__(self, b: float) -> Self: ...
    def __isub__(self, b: float) -> Self: ...  # type: ignore[misc] # __sub__ and __isub__ are incompatible
    def __radd__(self, b: float) -> Epoch: ...
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    def __eq__(self, b: float | Epoch) -> bool: ...  # type: ignore[override]
    def __ne__(self, b: float | Epoch) -> bool: ...  # type: ignore[override]
    def __lt__(self, b: float | Epoch) -> bool: ...
    def __ge__(self, b: float | Epoch) -> bool: ...
    def __gt__(self, b: float | Epoch) -> bool: ...
    def __le__(self, b: float | Epoch) -> bool: ...

JDE2000: Epoch

def main() -> None: ...



from asyncio import AbstractEventLoop
from collections.abc import Callable
from logging import Logger
from typing_extensions import Self

from ..connection import Parameters
from .base_connection import BaseConnection
from .utils import connection_workflow, io_services_utils, nbio_interface

LOGGER: Logger

class AsyncioConnection(BaseConnection):
    def __init__(
        self,
        parameters: Parameters | None = None,
        on_open_callback: Callable[[Self], object] | None = None,
        on_open_error_callback: Callable[[Self, BaseException], object] | None = None,
        on_close_callback: Callable[[Self, BaseException], object] | None = None,
        custom_ioloop: AbstractEventLoop | None = None,
        internal_connection_workflow: bool = True,
    ) -> None: ...
    @classmethod
    def create_connection(
        cls,
        connection_configs,
        on_done,
        custom_ioloop: AbstractEventLoop | None = None,
        workflow: connection_workflow.AbstractAMQPConnectionWorkflow | None = None,
    ): ...

class _AsyncioIOServicesAdapter(
    io_services_utils.SocketConnectionMixin,
    io_services_utils.StreamingConnectionMixin,
    nbio_interface.AbstractIOServices,
    nbio_interface.AbstractFileDescriptorServices,
):
    def __init__(self, loop: AbstractEventLoop | None = None) -> None: ...
    def get_native_ioloop(self): ...
    def close(self) -> None: ...
    def run(self) -> None: ...
    def stop(self) -> None: ...
    def add_callback_threadsafe(self, callback) -> None: ...
    def call_later(self, delay, callback): ...
    def getaddrinfo(self, host, port, on_done, family: int = 0, socktype: int = 0, proto: int = 0, flags: int = 0): ...
    def set_reader(self, fd, on_readable) -> None: ...
    def remove_reader(self, fd): ...
    def set_writer(self, fd, on_writable) -> None: ...
    def remove_writer(self, fd): ...

class _TimerHandle(nbio_interface.AbstractTimerReference):
    def __init__(self, handle) -> None: ...
    def cancel(self) -> None: ...

class _AsyncioIOReference(nbio_interface.AbstractIOReference):
    def __init__(self, future, on_done) -> None: ...
    def cancel(self): ...



from collections.abc import Generator
from typing import Any

def pycodestyle_logical(
    blank_before: Any,
    blank_lines: Any,
    checker_state: Any,
    hang_closing: Any,
    indent_char: Any,
    indent_level: Any,
    indent_size: Any,
    line_number: Any,
    lines: Any,
    logical_line: Any,
    max_doc_length: Any,
    noqa: Any,
    previous_indent_level: Any,
    previous_logical: Any,
    previous_unindented_logical_line: Any,
    tokens: Any,
    verbose: Any,
) -> Generator[tuple[int, str], None, None]: ...
def pycodestyle_physical(
    indent_char: Any,
    line_number: Any,
    lines: Any,
    max_line_length: Any,
    multiline: Any,
    noqa: Any,
    physical_line: Any,
    total_lines: Any,
) -> Generator[tuple[int, str], None, None]: ...



from _typeshed import Incomplete

from braintree.attribute_getter import AttributeGetter

text_type = str
raw_type = bytes

class Resource(AttributeGetter):
    @staticmethod
    def verify_keys(params, signature) -> None: ...
    gateway: Incomplete
    def __init__(self, gateway, attributes) -> None: ...



from _typeshed import Incomplete
from typing import ClassVar

class Behavior:
    name: str
    description: str
    versionString: str
    knownChildren: ClassVar[dict[str, tuple[int, int | None, int | None]]]
    quotedPrintable: bool
    defaultBehavior: Incomplete
    hasNative: bool
    isComponent: bool
    allowGroup: bool
    forceUTC: bool
    sortFirst: Incomplete
    @classmethod
    def validate(cls, obj, raiseException: bool = False, complainUnrecognized: bool = False) -> bool: ...
    @classmethod
    def lineValidate(cls, line, raiseException, complainUnrecognized): ...
    @classmethod
    def decode(cls, line) -> None: ...
    @classmethod
    def encode(cls, line) -> None: ...
    @classmethod
    def transformToNative(cls, obj): ...
    @classmethod
    def transformFromNative(cls, obj) -> None: ...
    @classmethod
    def generateImplicitParameters(cls, obj) -> None: ...
    @classmethod
    def serialize(cls, obj, buf, lineLength, validate: bool = True, *args, **kwargs): ...
    @classmethod
    def valueRepr(cls, line): ...



from _typeshed import Incomplete
from collections.abc import Iterable

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["is_bipartite", "is_bipartite_node_set", "color", "sets", "density", "degrees"]

@_dispatchable
def color(G: Graph[_Node]) -> dict[Incomplete, Incomplete]: ...
@_dispatchable
def is_bipartite(G: Graph[_Node]) -> bool: ...
@_dispatchable
def is_bipartite_node_set(G: Graph[_Node], nodes: Iterable[Incomplete]) -> bool: ...
@_dispatchable
def sets(G: Graph[_Node], top_nodes: Iterable[Incomplete] | None = None) -> tuple[set[Incomplete], set[Incomplete]]: ...
@_dispatchable
def density(B: Graph[_Node], nodes) -> float: ...
@_dispatchable
def degrees(B: Graph[_Node], nodes, weight: str | None = None) -> tuple[Incomplete, Incomplete]: ...



import asyncio
import sys


def get_set(loop: asyncio.BaseEventLoop) -> None:
    loop.set_task_factory(loop.get_task_factory())


if sys.version_info >= (3, 12):

    def eager(loop: asyncio.BaseEventLoop) -> None:
        loop.set_task_factory(asyncio.eager_task_factory)
        loop.set_task_factory(asyncio.create_eager_task_factory(asyncio.Task))



from collections.abc import Callable, Generator
from re import Pattern
from typing import Final, Literal, overload
from typing_extensions import TypeAlias

from ._types import ErrorCorrect, MaskPattern
from .base import RSBlock as RSBlock

MODE_NUMBER: Final[int] = 1
MODE_ALPHA_NUM: Final[int] = 2
MODE_8BIT_BYTE: Final[int] = 4
MODE_KANJI: Final[int] = 8

_MODE: TypeAlias = Literal[1, 2, 4, 8]

MODE_SIZE_SMALL: Final[dict[_MODE, int]]
MODE_SIZE_MEDIUM: Final[dict[_MODE, int]]
MODE_SIZE_LARGE: Final[dict[_MODE, int]]

ALPHA_NUM: Final[bytes]
RE_ALPHA_NUM: Final[Pattern[bytes]]
NUMBER_LENGTH: Final[dict[int, int]]
PATTERN_POSITION_TABLE: Final[list[list[int]]]
G15: Final[int]
G18: Final[int]
G15_MASK: Final[int]
PAD0: Final[int]
PAD1: Final[int]
BIT_LIMIT_TABLE: Final[list[list[int]]]

# In the implementation, MODE_KANJI is not accepted in all places
_SupportedMode: TypeAlias = Literal[1, 2, 4]

def BCH_type_info(data: int) -> int: ...
def BCH_type_number(data: int) -> int: ...
def BCH_digit(data: int) -> int: ...
def pattern_position(version: int) -> list[int]: ...
def mask_func(pattern: MaskPattern) -> Callable[[int, int], bool]: ...
def mode_sizes_for_version(version: int) -> dict[_MODE, int]: ...
def length_in_bits(mode: _MODE, version: int) -> int: ...
def check_version(version: int) -> None: ...
def lost_point(modules: list[list[bool | None]]) -> int: ...
def optimal_data_chunks(data: str | bytes, minimum: int = 4) -> Generator[QRData, None, None]: ...
def to_bytestring(data: str | bytes) -> bytes: ...
def optimal_mode(data: bytes) -> _SupportedMode: ...

class QRData:
    mode: _SupportedMode
    data: bytes
    @overload
    def __init__(self, data: bytes | str, mode: _SupportedMode | None = None, check_data: Literal[True] = True) -> None: ...
    @overload
    def __init__(self, data: bytes, mode: _SupportedMode | None = None, *, check_data: Literal[False]) -> None: ...
    @overload
    def __init__(self, data: bytes, mode: _SupportedMode | None, check_data: Literal[False]) -> None: ...
    def __len__(self) -> int: ...
    def write(self, buffer: BitBuffer) -> None: ...

class BitBuffer:
    buffer: list[int]
    length: int
    def __init__(self) -> None: ...
    def get(self, index: int) -> bool: ...
    def put(self, num: int, length: int) -> None: ...
    def __len__(self) -> int: ...
    def put_bit(self, bit: bool) -> None: ...

def create_bytes(buffer: BitBuffer, rs_blocks: list[RSBlock]) -> list[int]: ...
def create_data(version: int, error_correction: ErrorCorrect, data_list: list[QRData]) -> list[int]: ...



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import asyncio
import calendar
import contextlib
import datetime
import enum
import functools
import io
import os
import re
import shutil
import subprocess
import sys
import tarfile
import textwrap
import urllib.parse
import zipfile
from collections.abc import Callable, Iterator, Mapping, Sequence
from dataclasses import dataclass, field
from http import HTTPStatus
from pathlib import Path
from typing import Annotated, Any, ClassVar, Literal, NamedTuple, TypedDict, TypeVar
from typing_extensions import Self, TypeAlias

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

import aiohttp
import packaging.version
import tomlkit
from packaging.specifiers import Specifier
from termcolor import colored

from ts_utils.metadata import ObsoleteMetadata, StubMetadata, read_metadata, update_metadata
from ts_utils.paths import PYRIGHT_CONFIG, STUBS_PATH, distribution_path

TYPESHED_OWNER = "python"
TYPESHED_API_URL = f"https://api.github.com/repos/{TYPESHED_OWNER}/typeshed"

STUBSABOT_LABEL = "bot: stubsabot"

POLICY_MONTHS_DELTA = 6


class ActionLevel(enum.IntEnum):
    def __new__(cls, value: int, doc: str) -> Self:
        member = int.__new__(cls, value)
        member._value_ = value
        member.__doc__ = doc
        return member

    @classmethod
    def from_cmd_arg(cls, cmd_arg: str) -> ActionLevel:
        try:
            return cls[cmd_arg]
        except KeyError:
            raise argparse.ArgumentTypeError(f'Argument must be one of "{list(cls.__members__)}"') from None

    nothing = 0, "make no changes"
    local = 1, "make changes that affect local repo"
    fork = 2, "make changes that affect remote repo, but won't open PRs against upstream"
    everything = 3, "do everything, e.g. open PRs"


@dataclass
class PypiReleaseDownload:
    distribution: str
    url: str
    packagetype: Annotated[str, "Should hopefully be either 'bdist_wheel' or 'sdist'"]
    filename: str
    version: packaging.version.Version
    upload_date: datetime.datetime


VersionString: TypeAlias = str
ReleaseDownload: TypeAlias = dict[str, Any]


def _best_effort_version(version: VersionString) -> packaging.version.Version:
    try:
        return packaging.version.Version(version)
    except packaging.version.InvalidVersion:
        # packaging.version.Version no longer parses legacy versions
        try:
            return packaging.version.Version(version.replace("-", "+"))
        except packaging.version.InvalidVersion:
            return packaging.version.Version("0")


@dataclass
class PypiInfo:
    distribution: str
    pypi_root: str
    releases: dict[VersionString, list[ReleaseDownload]] = field(repr=False)
    info: dict[str, Any] = field(repr=False)

    def get_release(self, *, version: VersionString) -> PypiReleaseDownload:
        # prefer wheels, since it's what most users will get / it's pretty easy to mess up MANIFEST
        release_info = sorted(self.releases[version], key=lambda x: bool(x["packagetype"] == "bdist_wheel"))[-1]
        return PypiReleaseDownload(
            distribution=self.distribution,
            url=release_info["url"],
            packagetype=release_info["packagetype"],
            filename=release_info["filename"],
            version=packaging.version.Version(version),
            upload_date=datetime.datetime.fromisoformat(release_info["upload_time"]),
        )

    def get_latest_release(self) -> PypiReleaseDownload:
        return self.get_release(version=self.info["version"])

    def releases_in_descending_order(self) -> Iterator[PypiReleaseDownload]:
        for version in sorted(self.releases, key=_best_effort_version, reverse=True):
            yield self.get_release(version=version)


async def fetch_pypi_info(distribution: str, session: aiohttp.ClientSession) -> PypiInfo:
    # Cf. # https://warehouse.pypa.io/api-reference/json.html#get--pypi--project_name--json
    pypi_root = f"https://pypi.org/pypi/{urllib.parse.quote(distribution)}"
    async with session.get(f"{pypi_root}/json") as response:
        response.raise_for_status()
        j = await response.json()
        return PypiInfo(distribution=distribution, pypi_root=pypi_root, releases=j["releases"], info=j["info"])


@dataclass
class Update:
    distribution: str
    old_version_spec: Specifier
    new_version_spec: Specifier
    links: dict[str, str]
    diff_analysis: DiffAnalysis | None

    def __str__(self) -> str:
        return f"{colored('updating', 'yellow')} from '{self.old_version_spec}' to '{self.new_version_spec}'"

    @property
    def new_version(self) -> str:
        if self.new_version_spec.operator == "==":
            return str(self.new_version_spec)[2:]
        else:
            return str(self.new_version_spec)


@dataclass
class Obsolete:
    distribution: str
    obsolete_since_version: str
    obsolete_since_date: datetime.datetime
    links: dict[str, str]

    def __str__(self) -> str:
        return f"{colored('marking as obsolete', 'yellow')} since {self.obsolete_since_version!r}"


@dataclass
class Remove:
    distribution: str
    reason: Literal["ships py.typed file", "unmaintained"]
    links: dict[str, str]

    def __str__(self) -> str:
        return f"{colored('removing', 'yellow')} ({self.reason})"


@dataclass
class NoUpdate:
    distribution: str
    reason: Literal["obsolete", "no longer updated", "up to date"]

    def __str__(self) -> str:
        return f"{colored('skipping', 'green')} ({self.reason})"


@dataclass
class Error:
    distribution: str
    message: str

    def __str__(self) -> str:
        return f"{colored('error', 'red')} ({self.message})"


_T = TypeVar("_T")


async def with_extracted_archive(
    release_to_download: PypiReleaseDownload,
    *,
    session: aiohttp.ClientSession,
    handler: Callable[[zipfile.ZipFile | tarfile.TarFile], _T],
) -> _T:
    async with session.get(release_to_download.url) as response:
        body = io.BytesIO(await response.read())

    packagetype = release_to_download.packagetype
    if packagetype == "bdist_wheel":
        assert release_to_download.filename.endswith(".whl")
        with zipfile.ZipFile(body) as zf:
            return handler(zf)
    elif packagetype == "sdist":
        # sdist defaults to `.tar.gz` on Linux and to `.zip` on Windows:
        # https://docs.python.org/3.11/distutils/sourcedist.html
        if release_to_download.filename.endswith(".tar.gz"):
            with tarfile.open(fileobj=body, mode="r:gz") as zf:
                return handler(zf)
        elif release_to_download.filename.endswith(".zip"):
            with zipfile.ZipFile(body) as zf:
                return handler(zf)
        else:
            raise AssertionError(f"Package file {release_to_download.filename!r} does not end with '.tar.gz' or '.zip'")
    else:
        raise AssertionError(f"Unknown package type for {release_to_download.distribution}: {packagetype!r}")


def all_py_files_in_source_are_in_py_typed_dirs(source: zipfile.ZipFile | tarfile.TarFile) -> bool:
    py_typed_dirs: list[Path] = []
    all_python_files: list[Path] = []
    py_file_suffixes = {".py", ".pyi"}

    # Obtain an iterator over all files in the zipfile/tarfile.
    # Filter out empty __init__.py files: this reduces false negatives
    # in cases like this:
    #
    # repo_root/
    # ├─ src/
    # |  ├─ pkg/
    # |  |  ├─ __init__.py          <-- This file is empty
    # |  |  ├─ subpkg/
    # |  |  |  ├─ __init__.py
    # |  |  |  ├─ libraryfile.py
    # |  |  |  ├─ libraryfile2.py
    # |  |  |  ├─ py.typed
    if isinstance(source, zipfile.ZipFile):
        path_iter = (
            Path(zip_info.filename)
            for zip_info in source.infolist()
            if ((not zip_info.is_dir()) and not (Path(zip_info.filename).name == "__init__.py" and zip_info.file_size == 0))
        )
    else:
        path_iter = (
            Path(tar_info.path)
            for tar_info in source
            if (tar_info.isfile() and not (Path(tar_info.name).name == "__init__.py" and tar_info.size == 0))
        )

    for path in path_iter:
        if path.suffix in py_file_suffixes:
            all_python_files.append(path)
        elif path.name == "py.typed":
            py_typed_dirs.append(path.parent)

    if not py_typed_dirs:
        return False
    if not all_python_files:
        return False

    for path in all_python_files:
        if not any(py_typed_dir in path.parents for py_typed_dir in py_typed_dirs):
            return False
    return True


async def release_contains_py_typed(release_to_download: PypiReleaseDownload, *, session: aiohttp.ClientSession) -> bool:
    return await with_extracted_archive(release_to_download, session=session, handler=all_py_files_in_source_are_in_py_typed_dirs)


async def find_first_release_with_py_typed(pypi_info: PypiInfo, *, session: aiohttp.ClientSession) -> PypiReleaseDownload | None:
    """If the latest release is py.typed, return the first release that included a py.typed file.

    If the latest release is not py.typed, return None.
    """
    release_iter = (release for release in pypi_info.releases_in_descending_order() if not release.version.is_prerelease)
    latest_release = next(release_iter)
    # If the latest release is not py.typed, assume none are.
    if not (await release_contains_py_typed(latest_release, session=session)):
        return None

    first_release_with_py_typed = latest_release
    while await release_contains_py_typed(release := next(release_iter), session=session):
        first_release_with_py_typed = release
    return first_release_with_py_typed


def get_updated_version_spec(spec: Specifier, version: packaging.version.Version) -> Specifier:
    """
    Given the old specifier and an updated version, returns an updated specifier that has the
    specificity of the old specifier, but matches the updated version.

    For example:
    spec="1", version="1.2.3" -> "1.2.3"
    spec="1.0.1", version="1.2.3" -> "1.2.3"
    spec="1.*", version="1.2.3" -> "1.*"
    spec="1.*", version="2.3.4" -> "2.*"
    spec="1.1.*", version="1.2.3" -> "1.2.*"
    spec="1.1.1.*", version="1.2.3" -> "1.2.3.*"
    spec="~=1.0.1", version="1.0.3" -> "~=1.0.3"
    spec="~=1.0.1", version="1.1.0" -> "~=1.1.0"
    """
    if spec.operator == "==" and spec.version.endswith(".*"):
        specificity = spec.version.count(".") if spec.version.removesuffix(".*") else 0
        rounded_version = version.base_version.split(".")[:specificity]
        rounded_version.extend(["0"] * (specificity - len(rounded_version)))
        updated_spec = Specifier("==" + ".".join(rounded_version) + ".*")
    elif spec.operator == "==":
        updated_spec = Specifier(f"=={version}")
    elif spec.operator == "~=":
        updated_spec = Specifier(f"~={version}")
    else:
        raise ValueError(f"Unsupported version operator: {spec.operator}")
    assert version in updated_spec, f"{version} not in {updated_spec}"
    return updated_spec


@functools.cache
def get_github_api_headers() -> Mapping[str, str]:
    headers = {"Accept": "application/vnd.github.v3+json"}
    secret = os.environ.get("GITHUB_TOKEN")
    if secret is not None:
        headers["Authorization"] = f"token {secret}" if secret.startswith("ghp") else f"Bearer {secret}"
    return headers


GitHost: TypeAlias = Literal["github", "gitlab"]


@dataclass
class GitHostInfo:
    host: GitHost
    repo_path: str
    tags: list[str] = field(repr=False)


async def get_host_repo_info(session: aiohttp.ClientSession, stub_info: StubMetadata) -> GitHostInfo | None:
    """
    If the project represented by `stub_info` is publicly hosted (e.g. on GitHub)
    return information regarding the project as it exists on the public host.

    Else, return None.
    """
    if not stub_info.upstream_repository:
        return None
    # We have various sanity checks for the upstream_repository field in ts_utils.metadata,
    # so no need to repeat all of them here
    split_url = urllib.parse.urlsplit(stub_info.upstream_repository)
    host = split_url.netloc.removesuffix(".com")
    if host not in ("github", "gitlab"):
        return None
    url_path = split_url.path.strip("/")
    assert len(Path(url_path).parts) == 2
    if host == "github":
        # https://docs.github.com/en/rest/git/tags
        info_url = f"https://api.github.com/repos/{url_path}/tags"
        headers = get_github_api_headers()
    else:
        assert host == "gitlab"
        # https://docs.gitlab.com/api/tags/
        project_id = urllib.parse.quote(url_path, safe="")
        info_url = f"https://gitlab.com/api/v4/projects/{project_id}/repository/tags"
        headers = None
    async with session.get(info_url, headers=headers) as response:
        if response.status == HTTPStatus.OK:
            # Conveniently both GitHub and GitLab use the same key name.
            tags = [tag["name"] for tag in await response.json()]
            return GitHostInfo(host=host, repo_path=url_path, tags=tags)  # type: ignore[arg-type]
    return None


class GitHostDiffInfo(NamedTuple):
    host: GitHost
    repo_path: str
    old_tag: str
    new_tag: str

    @property
    def diff_url(self) -> str:
        if self.host == "github":
            return f"https://github.com/{self.repo_path}/compare/{self.old_tag}...{self.new_tag}"
        else:
            assert self.host == "gitlab"
            return f"https://gitlab.com/{self.repo_path}/-/compare/{self.old_tag}...{self.new_tag}"


async def get_diff_info(
    session: aiohttp.ClientSession, stub_info: StubMetadata, pypi_version: packaging.version.Version
) -> GitHostDiffInfo | None:
    """Return a tuple giving info about the diff between two releases, if possible.

    Return `None` if the project isn't hosted on GitHub,
    or if a link pointing to the diff couldn't be found for any other reason.
    """
    host_info = await get_host_repo_info(session, stub_info)
    if host_info is None:
        return None

    versions_to_tags: dict[packaging.version.Version, str] = {}
    for tag_name in host_info.tags:
        # Some packages in typeshed have tag names
        # that are invalid to be passed to the Version() constructor,
        # e.g. v.1.4.2
        with contextlib.suppress(packaging.version.InvalidVersion):
            versions_to_tags[packaging.version.Version(tag_name)] = tag_name

    try:
        new_tag = versions_to_tags[pypi_version]
    except KeyError:
        return None

    try:
        old_version = max(version for version in versions_to_tags if version in stub_info.version_spec and version < pypi_version)
    except ValueError:
        return None
    else:
        old_tag = versions_to_tags[old_version]

    return GitHostDiffInfo(host=host_info.host, repo_path=host_info.repo_path, old_tag=old_tag, new_tag=new_tag)


FileStatus: TypeAlias = Literal["added", "modified", "removed", "renamed"]


class FileInfo(TypedDict):
    filename: str
    status: FileStatus
    additions: int
    deletions: int


def _plural_s(num: int, /) -> str:
    return "s" if num != 1 else ""


@dataclass(repr=False)
class DiffAnalysis:
    MAXIMUM_NUMBER_OF_FILES_TO_LIST: ClassVar[int] = 7
    py_files: list[FileInfo]
    py_files_stubbed_in_typeshed: list[FileInfo]

    @property
    def runtime_definitely_has_consistent_directory_structure_with_typeshed(self) -> bool:
        """
        If 0 .py files in the GitHub diff exist in typeshed's stubs,
        there's a possibility that the .py files might be found
        in a different directory at runtime.

        For example: pyopenssl has its .py files in the `src/OpenSSL/` directory at runtime,
        but in typeshed the stubs are in the `OpenSSL/` directory.
        """
        return bool(self.py_files_stubbed_in_typeshed)

    @functools.cached_property
    def public_files_added(self) -> Sequence[str]:
        def is_public(path: Path) -> bool:
            return not re.match(r"_[^_]", path.name) and not path.name.startswith("test_")

        return [file["filename"] for file in self.py_files if is_public(Path(file["filename"])) and file["status"] == "added"]

    @functools.cached_property
    def typeshed_files_deleted(self) -> Sequence[str]:
        return [file["filename"] for file in self.py_files_stubbed_in_typeshed if file["status"] == "removed"]

    @functools.cached_property
    def typeshed_files_modified(self) -> Sequence[str]:
        return [file["filename"] for file in self.py_files_stubbed_in_typeshed if file["status"] in {"modified", "renamed"}]

    @property
    def total_lines_added(self) -> int:
        return sum(file["additions"] for file in self.py_files)

    @property
    def total_lines_deleted(self) -> int:
        return sum(file["deletions"] for file in self.py_files)

    def _describe_files(self, *, verb: str, filenames: Sequence[str]) -> str:
        num_files = len(filenames)
        if num_files > 1:
            description = f"have been {verb}"
            # Don't list the filenames if there are *loads* of files
            if num_files <= self.MAXIMUM_NUMBER_OF_FILES_TO_LIST:
                description += ": "
                description += ", ".join(f"`{filename}`" for filename in filenames)
            description += "."
            return description
        if num_files == 1:
            return f"has been {verb}: `{filenames[0]}`."
        return f"have been {verb}."

    def describe_public_files_added(self) -> str:
        num_files_added = len(self.public_files_added)
        analysis = f"{num_files_added} public Python file{_plural_s(num_files_added)} "
        analysis += self._describe_files(verb="added", filenames=self.public_files_added)
        return analysis

    def describe_typeshed_files_deleted(self) -> str:
        num_files_deleted = len(self.typeshed_files_deleted)
        analysis = f"{num_files_deleted} file{_plural_s(num_files_deleted)} included in typeshed's stubs "
        analysis += self._describe_files(verb="deleted", filenames=self.typeshed_files_deleted)
        return analysis

    def describe_typeshed_files_modified(self) -> str:
        num_files_modified = len(self.typeshed_files_modified)
        analysis = f"{num_files_modified} file{_plural_s(num_files_modified)} included in typeshed's stubs "
        analysis += self._describe_files(verb="modified or renamed", filenames=self.typeshed_files_modified)
        return analysis

    def __str__(self) -> str:
        data_points: list[str] = []
        if self.runtime_definitely_has_consistent_directory_structure_with_typeshed:
            data_points += [
                self.describe_public_files_added(),
                self.describe_typeshed_files_deleted(),
                self.describe_typeshed_files_modified(),
            ]
        data_points += [
            f"Total lines of Python code added: {self.total_lines_added}.",
            f"Total lines of Python code deleted: {self.total_lines_deleted}.",
        ]
        return "Stubsabot analysis of the diff between the two releases:\n - " + "\n - ".join(data_points)


async def analyze_github_diff(
    repo_path: str, distribution: str, old_tag: str, new_tag: str, *, session: aiohttp.ClientSession
) -> DiffAnalysis | None:
    url = f"https://api.github.com/repos/{repo_path}/compare/{old_tag}...{new_tag}"
    async with session.get(url, headers=get_github_api_headers()) as response:
        response.raise_for_status()
        json_resp: dict[str, list[FileInfo]] = await response.json()
        assert isinstance(json_resp, dict)
    # https://docs.github.com/en/rest/commits/commits#compare-two-commits
    py_files: list[FileInfo] = [file for file in json_resp["files"] if Path(file["filename"]).suffix == ".py"]
    stub_path = distribution_path(distribution)
    files_in_typeshed = set(stub_path.rglob("*.pyi"))
    py_files_stubbed_in_typeshed = [file for file in py_files if (stub_path / f"{file['filename']}i") in files_in_typeshed]
    return DiffAnalysis(py_files=py_files, py_files_stubbed_in_typeshed=py_files_stubbed_in_typeshed)


async def analyze_gitlab_diff(
    repo_path: str, distribution: str, old_tag: str, new_tag: str, *, session: aiohttp.ClientSession
) -> DiffAnalysis | None:
    # https://docs.gitlab.com/api/repositories/#compare-branches-tags-or-commits
    project_id = urllib.parse.quote(repo_path, safe="")
    url = f"https://gitlab.com/api/v4/projects/{project_id}/repository/compare?from={old_tag}&to={new_tag}"
    async with session.get(url) as response:
        response.raise_for_status()
        json_resp: dict[str, Any] = await response.json()
        assert isinstance(json_resp, dict)

    py_files: list[FileInfo] = []
    for file_diff in json_resp["diffs"]:
        filename = file_diff["new_path"]
        if Path(filename).suffix != ".py":
            continue
        status: FileStatus
        if file_diff["new_file"]:
            status = "added"
        elif file_diff["renamed_file"]:
            status = "renamed"
        elif file_diff["deleted_file"]:
            status = "removed"
        else:
            status = "modified"
        diff_lines = file_diff["diff"].splitlines()
        additions = sum(1 for ln in diff_lines if ln.startswith("+"))
        deletions = sum(1 for ln in diff_lines if ln.startswith("-"))
        py_files.append(FileInfo(filename=filename, status=status, additions=additions, deletions=deletions))

    stub_path = distribution_path(distribution)
    files_in_typeshed = set(stub_path.rglob("*.pyi"))
    py_files_stubbed_in_typeshed = [file for file in py_files if (stub_path / f"{file['filename']}i") in files_in_typeshed]
    return DiffAnalysis(py_files=py_files, py_files_stubbed_in_typeshed=py_files_stubbed_in_typeshed)


def _add_months(date: datetime.date, months: int) -> datetime.date:
    month = date.month - 1 + months
    year = date.year + month // 12
    month = month % 12 + 1
    day = min(date.day, calendar.monthrange(year, month)[1])
    return datetime.date(year, month, day)


def obsolete_more_than_n_months(since_date: datetime.date) -> bool:
    remove_date = _add_months(since_date, POLICY_MONTHS_DELTA)
    today = datetime.datetime.now(tz=datetime.timezone.utc).date()
    return remove_date <= today


def parse_no_longer_updated_from_archive(source: zipfile.ZipFile | tarfile.TarFile) -> bool:
    if isinstance(source, zipfile.ZipFile):
        try:
            file = source.open("METADATA.toml", "r")
        except KeyError:
            return False
    else:
        try:
            tarinfo = source.getmember("METADATA.toml")
            file = source.extractfile(tarinfo)  # type: ignore[assignment]
            if file is None:
                return False
        except KeyError:
            return False

    with file as f:
        toml_data: dict[str, object] = tomllib.load(f)

    no_longer_updated = toml_data.get("no_longer_updated", False)
    assert type(no_longer_updated) is bool
    return bool(no_longer_updated)


async def has_no_longer_updated_release(release_to_download: PypiReleaseDownload, *, session: aiohttp.ClientSession) -> bool:
    """
    Return `True` if the `no_longer_updated` field exists and the value is
    `True` in the `METADATA.toml` file of latest `types-{distribution}` pypi release.
    """
    return await with_extracted_archive(release_to_download, session=session, handler=parse_no_longer_updated_from_archive)


async def determine_action(distribution: str, session: aiohttp.ClientSession) -> Update | NoUpdate | Obsolete | Remove | Error:
    try:
        return await determine_action_no_error_handling(distribution, session)
    except Exception as exc:
        return Error(distribution, str(exc))


async def determine_action_no_error_handling(
    distribution: str, session: aiohttp.ClientSession
) -> Update | NoUpdate | Obsolete | Remove:
    stub_info = read_metadata(distribution)
    if stub_info.is_obsolete:
        assert type(stub_info.obsolete) is ObsoleteMetadata
        since_date = stub_info.obsolete.since_date

        if obsolete_more_than_n_months(since_date):
            pypi_info = await fetch_pypi_info(f"types-{stub_info.distribution}", session)
            latest_release = pypi_info.get_latest_release()
            links = {
                "Typeshed release": f"{pypi_info.pypi_root}",
                "Typeshed stubs": f"https://github.com/{TYPESHED_OWNER}/typeshed/tree/main/stubs/{stub_info.distribution}",
            }
            return Remove(stub_info.distribution, reason="ships py.typed file", links=links)
        else:
            return NoUpdate(stub_info.distribution, "obsolete")
    if stub_info.no_longer_updated:
        pypi_info = await fetch_pypi_info(f"types-{stub_info.distribution}", session)
        latest_release = pypi_info.get_latest_release()

        if await has_no_longer_updated_release(latest_release, session=session):
            links = {
                "Typeshed release": f"{pypi_info.pypi_root}",
                "Typeshed stubs": f"https://github.com/{TYPESHED_OWNER}/typeshed/tree/main/stubs/{stub_info.distribution}",
            }
            return Remove(stub_info.distribution, reason="unmaintained", links=links)
        else:
            return NoUpdate(stub_info.distribution, "no longer updated")

    pypi_info = await fetch_pypi_info(stub_info.distribution, session)
    latest_release = pypi_info.get_latest_release()
    latest_version = latest_release.version
    obsolete_since = await find_first_release_with_py_typed(pypi_info, session=session)
    if obsolete_since is None and latest_version in stub_info.version_spec:
        return NoUpdate(stub_info.distribution, "up to date")

    relevant_version = obsolete_since.version if obsolete_since else latest_version

    project_urls: dict[str, str] = pypi_info.info["project_urls"] or {}
    maybe_links: dict[str, str | None] = {
        "Release": f"{pypi_info.pypi_root}/{relevant_version}",
        "Homepage": project_urls.get("Homepage"),
        "Repository": stub_info.upstream_repository,
        "Typeshed stubs": f"https://github.com/{TYPESHED_OWNER}/typeshed/tree/main/stubs/{stub_info.distribution}",
        "Changelog": project_urls.get("Changelog") or project_urls.get("Changes") or project_urls.get("Change Log"),
    }
    links = {k: v for k, v in maybe_links.items() if v is not None}

    diff_info = await get_diff_info(session, stub_info, relevant_version)
    if diff_info is not None:
        links["Diff"] = diff_info.diff_url

    if obsolete_since:
        return Obsolete(
            stub_info.distribution,
            obsolete_since_version=str(obsolete_since.version),
            obsolete_since_date=obsolete_since.upload_date,
            links=links,
        )

    if diff_info is None:
        diff_analysis: DiffAnalysis | None = None
    else:
        analyze_diff = {"github": analyze_github_diff, "gitlab": analyze_gitlab_diff}[diff_info.host]
        diff_analysis = await analyze_diff(
            repo_path=diff_info.repo_path,
            distribution=distribution,
            old_tag=diff_info.old_tag,
            new_tag=diff_info.new_tag,
            session=session,
        )

    return Update(
        distribution=stub_info.distribution,
        old_version_spec=stub_info.version_spec,
        new_version_spec=get_updated_version_spec(stub_info.version_spec, latest_version),
        links=links,
        diff_analysis=diff_analysis,
    )


@functools.lru_cache
def get_origin_owner() -> str:
    output = subprocess.check_output(["git", "remote", "get-url", "origin"], text=True).strip()
    match = re.match(r"(git@github.com:|https://github.com/)(?P<owner>[^/]+)/(?P<repo>[^/\s]+)", output)
    assert match is not None, f"Couldn't identify origin's owner: {output!r}"
    assert match.group("repo").removesuffix(".git") == "typeshed", f'Unexpected repo: {match.group("repo")!r}'
    return match.group("owner")


async def create_or_update_pull_request(*, title: str, body: str, branch_name: str, session: aiohttp.ClientSession) -> None:
    fork_owner = get_origin_owner()

    async with session.post(
        f"{TYPESHED_API_URL}/pulls",
        json={"title": title, "body": body, "head": f"{fork_owner}:{branch_name}", "base": "main"},
        headers=get_github_api_headers(),
    ) as response:
        resp_json = await response.json()
        if response.status == HTTPStatus.CREATED:
            pr_number = resp_json["number"]
            assert isinstance(pr_number, int)
        elif response.status == HTTPStatus.UNPROCESSABLE_ENTITY and any(
            "A pull request already exists" in e.get("message", "") for e in resp_json.get("errors", [])
        ):
            pr_number = await update_existing_pull_request(title=title, body=body, branch_name=branch_name, session=session)
        else:
            response.raise_for_status()
            raise AssertionError(f"Unexpected response: {response.status}")
    await update_pull_request_label(pr_number=pr_number, session=session)


async def update_existing_pull_request(*, title: str, body: str, branch_name: str, session: aiohttp.ClientSession) -> int:
    fork_owner = get_origin_owner()

    # Find the existing PR
    async with session.get(
        f"{TYPESHED_API_URL}/pulls",
        params={"state": "open", "head": f"{fork_owner}:{branch_name}", "base": "main"},
        headers=get_github_api_headers(),
    ) as response:
        response.raise_for_status()
        resp_json = await response.json()
        assert len(resp_json) >= 1
        pr_number = resp_json[0]["number"]
        assert isinstance(pr_number, int)
    # Update the PR's title and body
    async with session.patch(
        f"{TYPESHED_API_URL}/pulls/{pr_number}", json={"title": title, "body": body}, headers=get_github_api_headers()
    ) as response:
        response.raise_for_status()
    return pr_number


async def update_pull_request_label(*, pr_number: int, session: aiohttp.ClientSession) -> None:
    # There is no pulls/.../labels endpoint, which is why we need to use the issues endpoint.
    async with session.post(
        f"{TYPESHED_API_URL}/issues/{pr_number}/labels", json={"labels": [STUBSABOT_LABEL]}, headers=get_github_api_headers()
    ) as response:
        response.raise_for_status()


def has_non_stubsabot_commits(branch: str) -> bool:
    assert not branch.startswith("origin/")
    try:
        # commits on origin/branch that are not on branch or are
        # patch equivalent to a commit on branch
        print(
            "[debugprint]",
            subprocess.check_output(
                ["git", "log", "--right-only", "--pretty=%an %s", "--cherry-pick", f"{branch}...origin/{branch}"]
            ),
        )
        print(
            "[debugprint]",
            subprocess.check_output(
                ["git", "log", "--right-only", "--pretty=%an", "--cherry-pick", f"{branch}...origin/{branch}"]
            ),
        )
        output = subprocess.check_output(
            ["git", "log", "--right-only", "--pretty=%an", "--cherry-pick", f"{branch}...origin/{branch}"],
            stderr=subprocess.DEVNULL,
        )
        return bool(set(output.splitlines()) - {b"stubsabot"})
    except subprocess.CalledProcessError:
        # origin/branch does not exist
        return False


def latest_commit_is_different_to_last_commit_on_origin(branch: str) -> bool:
    assert not branch.startswith("origin/")
    try:
        # https://www.git-scm.com/docs/git-range-diff
        # If the number of lines is >1,
        # it indicates that something about our commit is different to the last commit
        # (Could be the commit "content", or the commit message).
        commit_comparison = subprocess.run(
            ["git", "range-diff", f"origin/{branch}~1..origin/{branch}", "HEAD~1..HEAD"], check=True, capture_output=True
        )
        return len(commit_comparison.stdout.splitlines()) > 1
    except subprocess.CalledProcessError:
        # origin/branch does not exist
        return True


class RemoteConflictError(Exception):
    pass


def somewhat_safe_force_push(branch: str) -> None:
    if has_non_stubsabot_commits(branch):
        raise RemoteConflictError(f"origin/{branch} has non-stubsabot changes that are not on {branch}!")
    subprocess.check_call(["git", "push", "origin", branch, "--force"])


def normalize(name: str) -> str:
    # PEP 503 normalization
    return re.sub(r"[-_.]+", "-", name).lower()


# lock should be unnecessary, but can't hurt to enforce mutual exclusion
_repo_lock = asyncio.Lock()

BRANCH_PREFIX = "stubsabot"


def get_update_pr_body(update: Update, metadata: Mapping[str, Any]) -> str:
    body = "\n".join(f"{k}: {v}" for k, v in update.links.items())

    if update.diff_analysis is not None:
        body += f"\n\n{update.diff_analysis}"

    stubtest_settings: dict[str, Any] = metadata.get("tool", {}).get("stubtest", {})
    stubtest_will_run = not stubtest_settings.get("skip", False)
    if stubtest_will_run:
        body += textwrap.dedent(
            """

            If stubtest fails for this PR:
            - Leave this PR open (as a reminder, and to prevent stubsabot from opening another PR)
            - Fix stubtest failures in another PR, then close this PR

            Note that you will need to close and re-open the PR in order to trigger CI
            """
        )
    else:
        body += textwrap.dedent(
            f"""

            :warning: Review this PR manually, as stubtest is skipped in CI for {update.distribution}!
            Also check whether stubtest can be reenabled. :warning:
            """
        )
    return body


def remove_stubs(distribution: str) -> None:
    stub_path = distribution_path(distribution)
    target_path_prefix = f'"stubs/{distribution}'

    if stub_path.exists() and stub_path.is_dir():
        shutil.rmtree(stub_path)

    with PYRIGHT_CONFIG.open("r", encoding="UTF-8") as f:
        lines = f.readlines()

    lines = [line for line in lines if not line.lstrip().startswith(target_path_prefix)]

    with PYRIGHT_CONFIG.open("w", encoding="UTF-8") as f:
        f.writelines(lines)


async def suggest_typeshed_update(update: Update, session: aiohttp.ClientSession, action_level: ActionLevel) -> None:
    if action_level <= ActionLevel.nothing:
        return
    title = f"[stubsabot] Bump {update.distribution} to {update.new_version}"
    async with _repo_lock:
        branch_name = f"{BRANCH_PREFIX}/{normalize(update.distribution)}"
        subprocess.check_call(["git", "checkout", "-B", branch_name, "origin/main"])
        meta = update_metadata(update.distribution, version=update.new_version)
        body = get_update_pr_body(update, meta)
        subprocess.check_call(["git", "commit", "--all", "-m", f"{title}\n\n{body}"])
        if action_level <= ActionLevel.local:
            return
        if not latest_commit_is_different_to_last_commit_on_origin(branch_name):
            print(f"No pushing to origin required: origin/{branch_name} exists and requires no changes!")
            return
        somewhat_safe_force_push(branch_name)
        if action_level <= ActionLevel.fork:
            return

    await create_or_update_pull_request(title=title, body=body, branch_name=branch_name, session=session)


async def suggest_typeshed_obsolete(obsolete: Obsolete, session: aiohttp.ClientSession, action_level: ActionLevel) -> None:
    if action_level <= ActionLevel.nothing:
        return
    title = f"[stubsabot] Mark {obsolete.distribution} as obsolete since {obsolete.obsolete_since_version}"
    async with _repo_lock:
        branch_name = f"{BRANCH_PREFIX}/{normalize(obsolete.distribution)}"
        subprocess.check_call(["git", "checkout", "-B", branch_name, "origin/main"])
        obs_string = tomlkit.string(obsolete.obsolete_since_version)
        obs_string.comment(f"Released on {obsolete.obsolete_since_date.date().isoformat()}")
        update_metadata(obsolete.distribution, obsolete_since=obs_string)
        body = "\n".join(f"{k}: {v}" for k, v in obsolete.links.items())
        subprocess.check_call(["git", "commit", "--all", "-m", f"{title}\n\n{body}"])
        if action_level <= ActionLevel.local:
            return
        if not latest_commit_is_different_to_last_commit_on_origin(branch_name):
            print(f"No PR required: origin/{branch_name} exists and requires no changes!")
            return
        somewhat_safe_force_push(branch_name)
        if action_level <= ActionLevel.fork:
            return

    await create_or_update_pull_request(title=title, body=body, branch_name=branch_name, session=session)


async def suggest_typeshed_remove(remove: Remove, session: aiohttp.ClientSession, action_level: ActionLevel) -> None:
    if action_level <= ActionLevel.nothing:
        return
    title = f"[stubsabot] Remove {remove.distribution} as {remove.reason}"
    async with _repo_lock:
        branch_name = f"{BRANCH_PREFIX}/{normalize(remove.distribution)}"
        subprocess.check_call(["git", "checkout", "-B", branch_name, "origin/main"])
        remove_stubs(remove.distribution)
        body = "\n".join(f"{k}: {v}" for k, v in remove.links.items())
        subprocess.check_call(["git", "commit", "--all", "-m", f"{title}\n\n{body}"])
        if action_level <= ActionLevel.local:
            return
        if not latest_commit_is_different_to_last_commit_on_origin(branch_name):
            print(f"No pushing to origin required: origin/{branch_name} exists and requires no changes!")
            return
        somewhat_safe_force_push(branch_name)
        if action_level <= ActionLevel.fork:
            return

    await create_or_update_pull_request(title=title, body=body, branch_name=branch_name, session=session)


async def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--action-level",
        type=ActionLevel.from_cmd_arg,
        default=ActionLevel.everything,
        help="Limit actions performed to achieve dry runs for different levels of dryness",
    )
    parser.add_argument(
        "--action-count-limit",
        type=int,
        default=None,
        help="Limit number of actions performed and the remainder are logged. Useful for testing",
    )
    parser.add_argument("distributions", nargs="*", help="Distributions to update, default = all")
    args = parser.parse_args()

    if args.distributions:
        dists_to_update = args.distributions
    else:
        dists_to_update = [path.name for path in STUBS_PATH.iterdir()]

    if args.action_level > ActionLevel.nothing:
        subprocess.run(["git", "update-index", "--refresh"], capture_output=True, check=False)
        diff_result = subprocess.run(["git", "diff-index", "HEAD", "--name-only"], text=True, capture_output=True, check=False)
        if diff_result.returncode:
            print("Unexpected exception!")
            print(diff_result.stdout)
            print(diff_result.stderr)
            return diff_result.returncode
        if diff_result.stdout:
            changed_files = ", ".join(repr(line) for line in diff_result.stdout.split("\n") if line)
            print(f"Cannot run stubsabot, as uncommitted changes are present in {changed_files}!")
            return 1

    if args.action_level > ActionLevel.fork:
        if os.environ.get("GITHUB_TOKEN") is None:
            raise ValueError("GITHUB_TOKEN environment variable must be set")

    denylist = {"gdb"}  # gdb is not a pypi distribution

    original_branch = subprocess.run(
        ["git", "branch", "--show-current"], text=True, capture_output=True, check=True
    ).stdout.strip()

    if args.action_level >= ActionLevel.local:
        subprocess.check_call(["git", "fetch", "--prune", "--all"])

    error = False

    try:
        conn = aiohttp.TCPConnector(limit_per_host=10)
        async with aiohttp.ClientSession(connector=conn) as session:
            tasks = [
                asyncio.create_task(determine_action(distribution, session))
                for distribution in dists_to_update
                if distribution not in denylist
            ]

            action_count = 0
            for task in asyncio.as_completed(tasks):
                update = await task
                print(f"{update.distribution}... ", end="")
                print(update)

                if isinstance(update, NoUpdate):
                    continue
                if isinstance(update, Error):
                    error = True
                    continue

                if args.action_count_limit is not None and action_count >= args.action_count_limit:
                    print(colored("... but we've reached action count limit", "red"))
                    continue
                action_count += 1

                try:
                    if isinstance(update, Update):
                        await suggest_typeshed_update(update, session, action_level=args.action_level)
                        continue
                    if isinstance(update, Obsolete):
                        await suggest_typeshed_obsolete(update, session, action_level=args.action_level)
                        continue
                    # Redundant, but keeping for extra runtime validation
                    if isinstance(update, Remove):  # pyright: ignore[reportUnnecessaryIsInstance]
                        await suggest_typeshed_remove(update, session, action_level=args.action_level)
                        continue
                except RemoteConflictError as e:
                    print(colored(f"... but ran into {type(e).__qualname__}: {e}", "red"))
                    continue
                raise AssertionError
    finally:
        # if you need to cleanup, try:
        # git branch -D $(git branch --list 'stubsabot/*')
        if args.action_level >= ActionLevel.local and original_branch:
            subprocess.check_call(["git", "checkout", original_branch])

    return 1 if error else 0


if __name__ == "__main__":
    sys.exit(asyncio.run(main()))



from collections.abc import Generator, Iterable
from re import Pattern
from typing import ClassVar, Final, Literal

options: Final[str]

class smartchars:
    endash: ClassVar[str]
    emdash: ClassVar[str]
    ellipsis: ClassVar[str]
    apostrophe: ClassVar[str]
    quotes: ClassVar[dict[str, str | tuple[str, str, str, str]]]
    language: str
    def __init__(self, language: str = "en") -> None: ...

class RegularExpressions:
    START_SINGLE: ClassVar[Pattern[str]]
    START_DOUBLE: ClassVar[Pattern[str]]
    ADJACENT_1: ClassVar[Pattern[str]]
    ADJACENT_2: ClassVar[Pattern[str]]
    OPEN_SINGLE: ClassVar[Pattern[str]]
    OPEN_DOUBLE: ClassVar[Pattern[str]]
    DECADE: ClassVar[Pattern[str]]
    APOSTROPHE: ClassVar[Pattern[str]]
    OPENING_SECONDARY: ClassVar[Pattern[str]]
    CLOSING_SECONDARY: ClassVar[Pattern[str]]
    OPENING_PRIMARY: ClassVar[Pattern[str]]
    CLOSING_PRIMARY: ClassVar[Pattern[str]]

regexes: RegularExpressions
default_smartypants_attr: Final = "1"

def smartyPants(text: str, attr="1", language: str = "en") -> str: ...
def educate_tokens(text_tokens: Iterable[tuple[str, str]], attr="1", language: str = "en") -> Generator[str]: ...
def educateQuotes(text: str, language: str = "en") -> str: ...
def educateBackticks(text: str, language: str = "en") -> str: ...
def educateSingleBackticks(text: str, language: str = "en") -> str: ...
def educateDashes(text: str) -> str: ...
def educateDashesOldSchool(text: str) -> str: ...
def educateDashesOldSchoolInverted(text: str) -> str: ...
def educateEllipses(text: str) -> str: ...
def stupefyEntities(text: str, language: str = "en") -> str: ...
def processEscapes(text: str, restore: bool = False) -> str: ...
def tokenize(text: str) -> Generator[tuple[Literal["tag", "text"], str]]: ...



from _typeshed import Incomplete

from ..rest import RestClientOptions
from ..types import TimeoutType

class Tickets:
    domain: Incomplete
    protocol: Incomplete
    client: Incomplete
    def __init__(
        self,
        domain: str,
        token: str,
        telemetry: bool = True,
        timeout: TimeoutType = 5.0,
        protocol: str = "https",
        rest_options: RestClientOptions | None = None,
    ) -> None: ...
    def create_email_verification(self, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    async def create_email_verification_async(self, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    def create_pswd_change(self, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    async def create_pswd_change_async(self, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...



import datetime

from ..core import WesternCalendar

class Canada(WesternCalendar): ...

class LateFamilyDayMixin:
    def get_family_day(self, year: int, label: str = "Family Day") -> tuple[datetime.date, str]: ...

class VictoriaDayMixin:
    def get_victoria_day(self, year: int) -> tuple[datetime.date, str] | None: ...

class AugustCivicHolidayMixin:
    def get_civic_holiday(self, year: int, label: str = "Civic Holiday") -> tuple[datetime.date, str]: ...

class ThanksgivingMixin:
    def get_thanksgiving(self, year: int) -> tuple[datetime.date, str]: ...

class BoxingDayMixin:
    def get_boxing_day(self, year: int) -> list[tuple[datetime.date, str]]: ...

class StJeanBaptisteMixin:
    def get_st_jean(self, year: int) -> list[tuple[datetime.date, str]]: ...

class RemembranceDayShiftMixin:
    def get_remembrance_day(self, year: int) -> list[tuple[datetime.date, str]]: ...

class Ontario(BoxingDayMixin, ThanksgivingMixin, VictoriaDayMixin, LateFamilyDayMixin, AugustCivicHolidayMixin, Canada): ...
class Quebec(VictoriaDayMixin, StJeanBaptisteMixin, ThanksgivingMixin, Canada): ...

class BritishColumbia(VictoriaDayMixin, AugustCivicHolidayMixin, ThanksgivingMixin, Canada):
    def get_family_day(self, year: int) -> tuple[datetime.date, str]: ...

class Alberta(LateFamilyDayMixin, VictoriaDayMixin, ThanksgivingMixin, Canada): ...
class Saskatchewan(
    LateFamilyDayMixin, VictoriaDayMixin, RemembranceDayShiftMixin, AugustCivicHolidayMixin, ThanksgivingMixin, Canada
): ...
class Manitoba(LateFamilyDayMixin, VictoriaDayMixin, AugustCivicHolidayMixin, ThanksgivingMixin, Canada): ...
class NewBrunswick(AugustCivicHolidayMixin, Canada): ...
class NovaScotia(RemembranceDayShiftMixin, LateFamilyDayMixin, Canada): ...
class PrinceEdwardIsland(LateFamilyDayMixin, RemembranceDayShiftMixin, Canada): ...
class Newfoundland(Canada): ...
class Yukon(VictoriaDayMixin, ThanksgivingMixin, Canada): ...
class NorthwestTerritories(RemembranceDayShiftMixin, VictoriaDayMixin, ThanksgivingMixin, Canada): ...
class Nunavut(VictoriaDayMixin, ThanksgivingMixin, RemembranceDayShiftMixin, Canada): ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import sys
import typing

import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _CoordinatedTaskState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _CoordinatedTaskStateEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_CoordinatedTaskState.ValueType], builtins.type
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    TASKSTATE_UNSPECIFIED: _CoordinatedTaskState.ValueType  # 0
    """TASKSTATE_UNSPECIFIED is an invalid state such that indicates a bug."""
    TASKSTATE_UNINITIALIZED: _CoordinatedTaskState.ValueType  # 1
    """TASKSTATE_UNINITIALIZED is an agent-only state. While the agent is
    disconnected, the service has no way of knowing if the task is
    initialized/uninitialized.
    """
    TASKSTATE_DISCONNECTED: _CoordinatedTaskState.ValueType  # 2
    TASKSTATE_CONNECTED: _CoordinatedTaskState.ValueType  # 3
    TASKSTATE_ERROR: _CoordinatedTaskState.ValueType  # 4

class CoordinatedTaskState(_CoordinatedTaskState, metaclass=_CoordinatedTaskStateEnumTypeWrapper):
    """Represents the state of a remote worker"""

TASKSTATE_UNSPECIFIED: CoordinatedTaskState.ValueType  # 0
"""TASKSTATE_UNSPECIFIED is an invalid state such that indicates a bug."""
TASKSTATE_UNINITIALIZED: CoordinatedTaskState.ValueType  # 1
"""TASKSTATE_UNINITIALIZED is an agent-only state. While the agent is
disconnected, the service has no way of knowing if the task is
initialized/uninitialized.
"""
TASKSTATE_DISCONNECTED: CoordinatedTaskState.ValueType  # 2
TASKSTATE_CONNECTED: CoordinatedTaskState.ValueType  # 3
TASKSTATE_ERROR: CoordinatedTaskState.ValueType  # 4
global___CoordinatedTaskState = CoordinatedTaskState

@typing.final
class CoordinatedTask(google.protobuf.message.Message):
    """Represents a remote worker task, specified by job name and task id."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_NAME_FIELD_NUMBER: builtins.int
    TASK_ID_FIELD_NUMBER: builtins.int
    job_name: builtins.str
    task_id: builtins.int
    def __init__(self, *, job_name: builtins.str | None = ..., task_id: builtins.int | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["job_name", b"job_name", "task_id", b"task_id"]) -> None: ...

global___CoordinatedTask = CoordinatedTask

@typing.final
class CoordinationServiceError(google.protobuf.message.Message):
    """Status payload for all coordination service errors.
    Note: an empty proto may be set if the error is triggered by the task's own
    agent calls (i.e. not propagated by the service from another remote task).
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IS_REPORTED_ERROR_FIELD_NUMBER: builtins.int
    SOURCE_TASK_FIELD_NUMBER: builtins.int
    is_reported_error: builtins.bool
    """If true, error is reported via the agent API by the user (and not an
    internal service error).
    """
    @property
    def source_task(self) -> global___CoordinatedTask:
        """Denotes which task hit the error. If unset, the error originated from the
        same task that is processing this error.
        """

    def __init__(
        self, *, is_reported_error: builtins.bool | None = ..., source_task: global___CoordinatedTask | None = ...
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["source_task", b"source_task"]) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["is_reported_error", b"is_reported_error", "source_task", b"source_task"]
    ) -> None: ...

global___CoordinationServiceError = CoordinationServiceError

@typing.final
class CoordinatedTaskStateInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TASK_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    ERROR_CODE_FIELD_NUMBER: builtins.int
    ERROR_MESSAGE_FIELD_NUMBER: builtins.int
    ERROR_PAYLOAD_FIELD_NUMBER: builtins.int
    state: global___CoordinatedTaskState.ValueType
    error_code: builtins.int
    error_message: builtins.str
    @property
    def task(self) -> global___CoordinatedTask: ...
    @property
    def error_payload(self) -> global___CoordinationServiceError: ...
    def __init__(
        self,
        *,
        task: global___CoordinatedTask | None = ...,
        state: global___CoordinatedTaskState.ValueType | None = ...,
        error_code: builtins.int | None = ...,
        error_message: builtins.str | None = ...,
        error_payload: global___CoordinationServiceError | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["error_payload", b"error_payload", "task", b"task"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "error_code",
            b"error_code",
            "error_message",
            b"error_message",
            "error_payload",
            b"error_payload",
            "state",
            b"state",
            "task",
            b"task",
        ],
    ) -> None: ...

global___CoordinatedTaskStateInfo = CoordinatedTaskStateInfo

@typing.final
class DeviceInfo(google.protobuf.message.Message):
    """Placeholder message to be extended by other runtimes' device representations."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DEVICE_FIELD_NUMBER: builtins.int
    @property
    def device(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[google.protobuf.any_pb2.Any]: ...
    def __init__(self, *, device: collections.abc.Iterable[google.protobuf.any_pb2.Any] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["device", b"device"]) -> None: ...

global___DeviceInfo = DeviceInfo

@typing.final
class RegisterTaskRequest(google.protobuf.message.Message):
    """Request and response messages for registering a task to the cluster leader.
    A task is uniquely represented by its `job_name`, `task_id` and
    `incarnation`. Leader responds with its `incarnation` to identify a leader
    process.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INCARNATION_FIELD_NUMBER: builtins.int
    SOURCE_TASK_FIELD_NUMBER: builtins.int
    incarnation: builtins.int
    @property
    def source_task(self) -> global___CoordinatedTask: ...
    def __init__(self, *, incarnation: builtins.int | None = ..., source_task: global___CoordinatedTask | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["source_task", b"source_task"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["incarnation", b"incarnation", "source_task", b"source_task"]) -> None: ...

global___RegisterTaskRequest = RegisterTaskRequest

@typing.final
class RegisterTaskResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LEADER_INCARNATION_FIELD_NUMBER: builtins.int
    leader_incarnation: builtins.int
    def __init__(self, *, leader_incarnation: builtins.int | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["leader_incarnation", b"leader_incarnation"]) -> None: ...

global___RegisterTaskResponse = RegisterTaskResponse

@typing.final
class HeartbeatRequest(google.protobuf.message.Message):
    """Request and response messages for sending heartbeats."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INCARNATION_FIELD_NUMBER: builtins.int
    SOURCE_TASK_FIELD_NUMBER: builtins.int
    incarnation: builtins.int
    @property
    def source_task(self) -> global___CoordinatedTask: ...
    def __init__(self, *, incarnation: builtins.int | None = ..., source_task: global___CoordinatedTask | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["source_task", b"source_task"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["incarnation", b"incarnation", "source_task", b"source_task"]) -> None: ...

global___HeartbeatRequest = HeartbeatRequest

@typing.final
class HeartbeatResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LEADER_INCARNATION_FIELD_NUMBER: builtins.int
    leader_incarnation: builtins.int
    """If there are failures in cluster, use additional metadata in response to
    broadcast error code and message to other tasks.
    """
    def __init__(self, *, leader_incarnation: builtins.int | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["leader_incarnation", b"leader_incarnation"]) -> None: ...

global___HeartbeatResponse = HeartbeatResponse

@typing.final
class PollForErrorRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SOURCE_TASK_FIELD_NUMBER: builtins.int
    @property
    def source_task(self) -> global___CoordinatedTask: ...
    def __init__(self, *, source_task: global___CoordinatedTask | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["source_task", b"source_task"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["source_task", b"source_task"]) -> None: ...

global___PollForErrorRequest = PollForErrorRequest

@typing.final
class PollForErrorResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___PollForErrorResponse = PollForErrorResponse

@typing.final
class WaitForAllTasksRequest(google.protobuf.message.Message):
    """Request and response messages for waiting for all tasks."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SOURCE_TASK_FIELD_NUMBER: builtins.int
    DEVICE_INFO_FIELD_NUMBER: builtins.int
    @property
    def source_task(self) -> global___CoordinatedTask: ...
    @property
    def device_info(self) -> global___DeviceInfo:
        """All local device attributes on the request sender;"""

    def __init__(
        self, *, source_task: global___CoordinatedTask | None = ..., device_info: global___DeviceInfo | None = ...
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["device_info", b"device_info", "source_task", b"source_task"]
    ) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["device_info", b"device_info", "source_task", b"source_task"]) -> None: ...

global___WaitForAllTasksRequest = WaitForAllTasksRequest

@typing.final
class WaitForAllTasksResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LEADER_INCARNATION_FIELD_NUMBER: builtins.int
    DEVICE_INFO_FIELD_NUMBER: builtins.int
    leader_incarnation: builtins.int
    @property
    def device_info(self) -> global___DeviceInfo:
        """All devices in the cluster."""

    def __init__(
        self, *, leader_incarnation: builtins.int | None = ..., device_info: global___DeviceInfo | None = ...
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["device_info", b"device_info"]) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["device_info", b"device_info", "leader_incarnation", b"leader_incarnation"]
    ) -> None: ...

global___WaitForAllTasksResponse = WaitForAllTasksResponse

@typing.final
class ShutdownTaskRequest(google.protobuf.message.Message):
    """Request and response messages for disconnecting a task from the service."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SOURCE_TASK_FIELD_NUMBER: builtins.int
    @property
    def source_task(self) -> global___CoordinatedTask: ...
    def __init__(self, *, source_task: global___CoordinatedTask | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["source_task", b"source_task"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["source_task", b"source_task"]) -> None: ...

global___ShutdownTaskRequest = ShutdownTaskRequest

@typing.final
class ShutdownTaskResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___ShutdownTaskResponse = ShutdownTaskResponse

@typing.final
class ResetTaskRequest(google.protobuf.message.Message):
    """Request and response messages for resetting a task state in the service."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SOURCE_TASK_FIELD_NUMBER: builtins.int
    @property
    def source_task(self) -> global___CoordinatedTask: ...
    def __init__(self, *, source_task: global___CoordinatedTask | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["source_task", b"source_task"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["source_task", b"source_task"]) -> None: ...

global___ResetTaskRequest = ResetTaskRequest

@typing.final
class ResetTaskResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___ResetTaskResponse = ResetTaskResponse

@typing.final
class ReportErrorToTaskRequest(google.protobuf.message.Message):
    """Request and response messages for reporting errors to task."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ERROR_CODE_FIELD_NUMBER: builtins.int
    ERROR_MESSAGE_FIELD_NUMBER: builtins.int
    ERROR_PAYLOAD_FIELD_NUMBER: builtins.int
    error_code: builtins.int
    error_message: builtins.str
    @property
    def error_payload(self) -> global___CoordinationServiceError: ...
    def __init__(
        self,
        *,
        error_code: builtins.int | None = ...,
        error_message: builtins.str | None = ...,
        error_payload: global___CoordinationServiceError | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["error_payload", b"error_payload"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "error_code", b"error_code", "error_message", b"error_message", "error_payload", b"error_payload"
        ],
    ) -> None: ...

global___ReportErrorToTaskRequest = ReportErrorToTaskRequest

@typing.final
class ReportErrorToTaskResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___ReportErrorToTaskResponse = ReportErrorToTaskResponse

@typing.final
class ReportErrorToServiceRequest(google.protobuf.message.Message):
    """Request and response messages for reporting errors to service instance."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ERROR_CODE_FIELD_NUMBER: builtins.int
    ERROR_MESSAGE_FIELD_NUMBER: builtins.int
    ERROR_ORIGIN_FIELD_NUMBER: builtins.int
    error_code: builtins.int
    error_message: builtins.str
    @property
    def error_origin(self) -> global___CoordinatedTask: ...
    def __init__(
        self,
        *,
        error_code: builtins.int | None = ...,
        error_message: builtins.str | None = ...,
        error_origin: global___CoordinatedTask | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["error_origin", b"error_origin"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "error_code", b"error_code", "error_message", b"error_message", "error_origin", b"error_origin"
        ],
    ) -> None: ...

global___ReportErrorToServiceRequest = ReportErrorToServiceRequest

@typing.final
class ReportErrorToServiceResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___ReportErrorToServiceResponse = ReportErrorToServiceResponse

@typing.final
class GetTaskStateRequest(google.protobuf.message.Message):
    """Request and response messages for getting state of a remote task."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SOURCE_TASK_FIELD_NUMBER: builtins.int
    @property
    def source_task(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___CoordinatedTask]: ...
    def __init__(self, *, source_task: collections.abc.Iterable[global___CoordinatedTask] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["source_task", b"source_task"]) -> None: ...

global___GetTaskStateRequest = GetTaskStateRequest

@typing.final
class GetTaskStateResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TASK_STATE_FIELD_NUMBER: builtins.int
    @property
    def task_state(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___CoordinatedTaskStateInfo]: ...
    def __init__(self, *, task_state: collections.abc.Iterable[global___CoordinatedTaskStateInfo] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["task_state", b"task_state"]) -> None: ...

global___GetTaskStateResponse = GetTaskStateResponse

@typing.final
class KeyValueEntry(google.protobuf.message.Message):
    """Message for configuration key value.
    Key is structured like Unix file system, with multiple levels of directory
    names separated by the slash ('/') characters.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    key: builtins.str
    value: builtins.bytes
    def __init__(self, *, key: builtins.str | None = ..., value: builtins.bytes | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

global___KeyValueEntry = KeyValueEntry

@typing.final
class InsertKeyValueRequest(google.protobuf.message.Message):
    """Request and response messages for inserting configuration key-value data."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KV_FIELD_NUMBER: builtins.int
    ALLOW_OVERWRITE_FIELD_NUMBER: builtins.int
    allow_overwrite: builtins.bool
    @property
    def kv(self) -> global___KeyValueEntry: ...
    def __init__(self, *, kv: global___KeyValueEntry | None = ..., allow_overwrite: builtins.bool | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["kv", b"kv"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["allow_overwrite", b"allow_overwrite", "kv", b"kv"]) -> None: ...

global___InsertKeyValueRequest = InsertKeyValueRequest

@typing.final
class InsertKeyValueResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___InsertKeyValueResponse = InsertKeyValueResponse

@typing.final
class GetKeyValueRequest(google.protobuf.message.Message):
    """Request and response messages for getting configuration key-value data."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    key: builtins.str
    def __init__(self, *, key: builtins.str | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["key", b"key"]) -> None: ...

global___GetKeyValueRequest = GetKeyValueRequest

@typing.final
class GetKeyValueResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KV_FIELD_NUMBER: builtins.int
    @property
    def kv(self) -> global___KeyValueEntry: ...
    def __init__(self, *, kv: global___KeyValueEntry | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["kv", b"kv"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["kv", b"kv"]) -> None: ...

global___GetKeyValueResponse = GetKeyValueResponse

@typing.final
class TryGetKeyValueRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    key: builtins.str
    def __init__(self, *, key: builtins.str | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["key", b"key"]) -> None: ...

global___TryGetKeyValueRequest = TryGetKeyValueRequest

@typing.final
class TryGetKeyValueResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KV_FIELD_NUMBER: builtins.int
    @property
    def kv(self) -> global___KeyValueEntry: ...
    def __init__(self, *, kv: global___KeyValueEntry | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["kv", b"kv"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["kv", b"kv"]) -> None: ...

global___TryGetKeyValueResponse = TryGetKeyValueResponse

@typing.final
class GetKeyValueDirRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DIRECTORY_KEY_FIELD_NUMBER: builtins.int
    directory_key: builtins.str
    def __init__(self, *, directory_key: builtins.str | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["directory_key", b"directory_key"]) -> None: ...

global___GetKeyValueDirRequest = GetKeyValueDirRequest

@typing.final
class GetKeyValueDirResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DIRECTORY_KEY_FIELD_NUMBER: builtins.int
    KV_FIELD_NUMBER: builtins.int
    directory_key: builtins.str
    @property
    def kv(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyValueEntry]: ...
    def __init__(
        self, *, directory_key: builtins.str | None = ..., kv: collections.abc.Iterable[global___KeyValueEntry] | None = ...
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["directory_key", b"directory_key", "kv", b"kv"]) -> None: ...

global___GetKeyValueDirResponse = GetKeyValueDirResponse

@typing.final
class DeleteKeyValueRequest(google.protobuf.message.Message):
    """Request and response messages for deleting configuration key-value data.
    When is_directory is true, delete key-values recursively under `key`.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    IS_DIRECTORY_FIELD_NUMBER: builtins.int
    key: builtins.str
    is_directory: builtins.bool
    def __init__(self, *, key: builtins.str | None = ..., is_directory: builtins.bool | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["is_directory", b"is_directory", "key", b"key"]) -> None: ...

global___DeleteKeyValueRequest = DeleteKeyValueRequest

@typing.final
class DeleteKeyValueResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___DeleteKeyValueResponse = DeleteKeyValueResponse

@typing.final
class BarrierRequest(google.protobuf.message.Message):
    """Request and response messages for generic sync barriers."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BARRIER_ID_FIELD_NUMBER: builtins.int
    BARRIER_TIMEOUT_IN_MS_FIELD_NUMBER: builtins.int
    TASKS_FIELD_NUMBER: builtins.int
    SOURCE_TASK_FIELD_NUMBER: builtins.int
    barrier_id: builtins.str
    barrier_timeout_in_ms: builtins.int
    @property
    def tasks(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___CoordinatedTask]:
        """Denotes list of tasks that will wait for the barrier. If unspecified, it
        implies that the entire cluster is participating in the barrier.
        """

    @property
    def source_task(self) -> global___CoordinatedTask:
        """Task that is making the request."""

    def __init__(
        self,
        *,
        barrier_id: builtins.str | None = ...,
        barrier_timeout_in_ms: builtins.int | None = ...,
        tasks: collections.abc.Iterable[global___CoordinatedTask] | None = ...,
        source_task: global___CoordinatedTask | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["source_task", b"source_task"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "barrier_id",
            b"barrier_id",
            "barrier_timeout_in_ms",
            b"barrier_timeout_in_ms",
            "source_task",
            b"source_task",
            "tasks",
            b"tasks",
        ],
    ) -> None: ...

global___BarrierRequest = BarrierRequest

@typing.final
class BarrierResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___BarrierResponse = BarrierResponse

@typing.final
class CancelBarrierRequest(google.protobuf.message.Message):
    """Request and response messages for  cancelling generic sync barriers."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BARRIER_ID_FIELD_NUMBER: builtins.int
    SOURCE_TASK_FIELD_NUMBER: builtins.int
    barrier_id: builtins.str
    @property
    def source_task(self) -> global___CoordinatedTask:
        """Task that is making the request."""

    def __init__(self, *, barrier_id: builtins.str | None = ..., source_task: global___CoordinatedTask | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["source_task", b"source_task"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["barrier_id", b"barrier_id", "source_task", b"source_task"]) -> None: ...

global___CancelBarrierRequest = CancelBarrierRequest

@typing.final
class CancelBarrierResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___CancelBarrierResponse = CancelBarrierResponse



import datetime

from ..core import OrthodoxCalendar

class Romania(OrthodoxCalendar):
    def get_childrens_day(self, year: int) -> list[tuple[datetime.date, str]]: ...
    def get_liberation_day(self, year: int) -> list[tuple[datetime.date, str]]: ...



import datetime
import optparse
from collections.abc import Sequence
from typing import Final, Literal

from .base import Component

version: Final[str]

def change_tz(
    cal: Component,
    new_timezone: datetime._TzInfo | None,
    default: datetime._TzInfo | None,
    utc_only: bool = False,
    utc_tz: datetime._TzInfo | None = ...,
) -> None: ...
def show_timezones() -> None: ...
def convert_events(utc_only: bool, args: Sequence[str]) -> None: ...
def main() -> None: ...
def get_options() -> tuple[optparse.Values, Literal[False]] | tuple[optparse.Values, list[str]]: ...



from _typeshed import Incomplete
from collections.abc import Callable
from typing import Any
from typing_extensions import TypeAlias

from matplotlib.markers import MarkerStyle
from matplotlib.path import Path
from numpy.typing import ArrayLike
from pandas import Series
from seaborn._core.scales import Scale

RGBTuple: TypeAlias = tuple[float, float, float]
RGBATuple: TypeAlias = tuple[float, float, float, float]
ColorSpec: TypeAlias = RGBTuple | RGBATuple | str
DashPattern: TypeAlias = tuple[float, ...]
DashPatternWithOffset: TypeAlias = tuple[float, DashPattern | None]
MarkerPattern: TypeAlias = float | str | tuple[int, int, float] | list[tuple[float, float]] | Path | MarkerStyle
Mapping: TypeAlias = Callable[[ArrayLike], ArrayLike]

class Property:
    legend: bool
    normed: bool
    variable: Incomplete
    def __init__(self, variable: str | None = None) -> None: ...
    def default_scale(self, data: Series[Any]) -> Scale: ...
    def infer_scale(self, arg: Any, data: Series[Any]) -> Scale: ...
    def get_mapping(self, scale: Scale, data: Series[Any]) -> Mapping: ...
    def standardize(self, val: Any) -> Any: ...

class Coordinate(Property):
    legend: bool
    normed: bool

class IntervalProperty(Property):
    legend: bool
    normed: bool
    @property
    def default_range(self) -> tuple[float, float]: ...
    def infer_scale(self, arg: Any, data: Series[Any]) -> Scale: ...
    def get_mapping(self, scale: Scale, data: Series[Any]) -> Mapping: ...

class PointSize(IntervalProperty): ...

class LineWidth(IntervalProperty):
    @property
    def default_range(self) -> tuple[float, float]: ...

class EdgeWidth(IntervalProperty):
    @property
    def default_range(self) -> tuple[float, float]: ...

class Stroke(IntervalProperty): ...
class Alpha(IntervalProperty): ...
class Offset(IntervalProperty): ...

class FontSize(IntervalProperty):
    @property
    def default_range(self) -> tuple[float, float]: ...

class ObjectProperty(Property):
    legend: bool
    normed: bool
    null_value: Any
    def default_scale(self, data: Series[Any]) -> Scale: ...
    def infer_scale(self, arg: Any, data: Series[Any]) -> Scale: ...
    def get_mapping(self, scale: Scale, data: Series[Any]) -> Mapping: ...

class Marker(ObjectProperty):
    null_value: Incomplete
    def standardize(self, val: MarkerPattern) -> MarkerStyle: ...

class LineStyle(ObjectProperty):
    null_value: str
    def standardize(self, val: str | DashPattern) -> DashPatternWithOffset: ...

class TextAlignment(ObjectProperty):
    legend: bool

class HorizontalAlignment(TextAlignment): ...
class VerticalAlignment(TextAlignment): ...

class Color(Property):
    legend: bool
    normed: bool
    def standardize(self, val: ColorSpec) -> RGBTuple | RGBATuple: ...
    def infer_scale(self, arg: Any, data: Series[Any]) -> Scale: ...
    def get_mapping(self, scale: Scale, data: Series[Any]) -> Mapping: ...

class Fill(Property):
    legend: bool
    normed: bool
    def default_scale(self, data: Series[Any]) -> Scale: ...
    def infer_scale(self, arg: Any, data: Series[Any]) -> Scale: ...
    def standardize(self, val: Any) -> bool: ...
    def get_mapping(self, scale: Scale, data: Series[Any]) -> Mapping: ...

PROPERTY_CLASSES: dict[str, type[Property]]
PROPERTIES: dict[str, Property]



import numbers
import xml.etree.ElementTree as ET
from collections.abc import Iterable
from typing import ClassVar, Final, SupportsIndex, overload
from typing_extensions import Self

__docformat__: Final = "reStructuredText"
GLOBAL_ATTRIBUTES: Final[tuple[str, ...]]

class MathElement(ET.Element):
    nchildren: ClassVar[int | None]
    parent: MathElement | None
    def __init__(self, *children, **attributes: object) -> None: ...  # attributes is passed to self.a_str method
    @staticmethod
    def a_str(v: object) -> str: ...
    def set(self, key: str, value: object) -> None: ...  # value is passed to self.a_str method
    @overload  # type: ignore[override]
    def __setitem__(self, key: SupportsIndex, value: MathElement) -> None: ...
    @overload  # type: ignore[override]
    def __setitem__(self, key: slice, value: Iterable[MathElement]) -> None: ...
    def is_full(self) -> bool: ...
    def close(self) -> MathElement | None: ...
    def append(self, element: MathElement) -> Self: ...  # type: ignore[override]
    def extend(self, elements: Iterable[MathElement]) -> Self: ...  # type: ignore[override]
    def pop(self, index: int = -1): ...
    def in_block(self) -> bool: ...
    def indent_xml(self, space: str = "  ", level: int = 0) -> None: ...
    def unindent_xml(self) -> None: ...
    def toxml(self, encoding: str | None = None) -> str: ...

class MathRow(MathElement): ...

class MathSchema(MathElement):
    nchildren: ClassVar[int]
    switch: bool
    def __init__(self, *children, switch: bool = False, **kwargs) -> None: ...

class MathToken(MathElement):
    nchildren: ClassVar[int]
    text: str
    def __init__(self, text: str | numbers.Number, **attributes: object) -> None: ...

class math(MathRow): ...
class mtext(MathToken): ...
class mi(MathToken): ...
class mn(MathToken): ...
class mo(MathToken): ...

class mspace(MathElement):
    nchildren: ClassVar[int]

class mrow(MathRow):
    def transfer_attributes(self, other) -> None: ...

class mfrac(MathSchema): ...

class msqrt(MathRow):
    nchildren: ClassVar[int]

class mroot(MathSchema): ...
class mstyle(MathRow): ...
class merror(MathRow): ...

class menclose(MathRow):
    nchildren: ClassVar[int]

class mpadded(MathRow): ...

class mphantom(MathRow):
    nchildren: ClassVar[int]

class msub(MathSchema): ...
class msup(MathSchema): ...

class msubsup(MathSchema):
    nchildren: ClassVar[int]

class munder(msub): ...
class mover(msup): ...
class munderover(msubsup): ...
class mtable(MathElement): ...
class mtr(MathRow): ...
class mtd(MathRow): ...



"""Helpers for generating target-specific JSX AST nodes."""

from __future__ import annotations

import ast as ast3
from typing import TYPE_CHECKING, cast

import jaclang.pycore.unitree as uni

if TYPE_CHECKING:
    from jaclang.compiler.passes.ecmascript.esast_gen_pass import EsastGenPass
    from jaclang.compiler.passes.ecmascript.estree import (
        Expression,
        Property,
        SpreadElement,
    )
    from jaclang.pycore.passes.pyast_gen_pass import PyastGenPass


class EsJsxProcessor:
    """Generate ESTree structures for JSX nodes."""

    def __init__(self, pass_ref: EsastGenPass) -> None:
        self.pass_ref = pass_ref
        # Import estree at runtime to access AST node classes
        from jaclang.compiler.passes.ecmascript import estree as es

        self.es = es

    def element(self, node: uni.JsxElement) -> Expression:
        """Process JSX element into __jacJsx(tag, props, children) call."""
        es = self.es
        if node.is_fragment or not node.name:
            tag_expr: Expression = self.pass_ref.sync_loc(
                es.Literal(value=None), jac_node=node
            )
        else:
            tag_expr = cast(
                "Expression",
                node.name.gen.es_ast
                if node.name.gen.es_ast
                else self.pass_ref.sync_loc(es.Literal(value=None), jac_node=node.name),
            )

        attributes = node.attributes or []
        has_spread = any(
            isinstance(attr, uni.JsxSpreadAttribute) for attr in attributes
        )
        if not attributes:
            props_expr: Expression = self.pass_ref.sync_loc(
                es.ObjectExpression(properties=[]), jac_node=node
            )
        elif has_spread:
            segments: list[Expression] = []
            for attr in attributes:
                if isinstance(attr, uni.JsxSpreadAttribute):
                    exp = attr.gen.es_ast
                    if isinstance(exp, es.Expression):
                        segments.append(exp)
                elif isinstance(attr, uni.JsxNormalAttribute):
                    prop = attr.gen.es_ast
                    if isinstance(prop, es.Property):
                        segments.append(
                            self.pass_ref.sync_loc(
                                es.ObjectExpression(properties=[prop]), jac_node=attr
                            )
                        )
            if segments:
                assign_member = self.pass_ref.sync_loc(
                    es.MemberExpression(
                        object=self.pass_ref.sync_loc(
                            es.Identifier(name="Object"), jac_node=node
                        ),
                        property=self.pass_ref.sync_loc(
                            es.Identifier(name="assign"), jac_node=node
                        ),
                        computed=False,
                        optional=False,
                    ),
                    jac_node=node,
                )
                props_expr = self.pass_ref.sync_loc(
                    es.CallExpression(
                        callee=assign_member,
                        arguments=[
                            self.pass_ref.sync_loc(
                                es.ObjectExpression(properties=[]), jac_node=node
                            ),
                            *segments,
                        ],
                    ),
                    jac_node=node,
                )
            else:
                props_expr = self.pass_ref.sync_loc(
                    es.ObjectExpression(properties=[]), jac_node=node
                )
        else:
            properties: list[Property] = []
            for attr in attributes:
                prop = attr.gen.es_ast
                if isinstance(prop, es.Property):
                    properties.append(prop)
            props_expr = self.pass_ref.sync_loc(
                es.ObjectExpression(properties=properties), jac_node=node
            )

        children_elements: list[Expression | SpreadElement | None] = []
        for child in node.children or []:
            child_expr = child.gen.es_ast
            if child_expr is None:
                continue
            if isinstance(child_expr, list):
                # JSX child expressions in list form are Expression types
                children_elements.extend(
                    cast("Expression", expr) for expr in child_expr
                )
            else:
                # Child expressions are always Expression types
                children_elements.append(cast("Expression", child_expr))
        children_expr = self.pass_ref.sync_loc(
            es.ArrayExpression(elements=children_elements), jac_node=node
        )

        call_expr = self.pass_ref.sync_loc(
            es.CallExpression(
                callee=self.pass_ref.sync_loc(
                    es.Identifier(name="__jacJsx"), jac_node=node
                ),
                arguments=[tag_expr, props_expr, children_expr],
            ),
            jac_node=node,
        )
        return call_expr

    def element_name(self, node: uni.JsxElementName) -> Expression:
        """Process JSX element name."""
        es = self.es
        if not node.parts:
            expr = self.pass_ref.sync_loc(es.Literal(value=None), jac_node=node)
        else:
            parts = [part.value for part in node.parts]
            first = parts[0]
            # If there are multiple parts (dotted name like motion.section),
            # always treat as member expression, not string literal
            # This allows imported objects like motion.section to work correctly
            if len(parts) > 1:
                # Multiple parts - create member expression chain
                expr = self.pass_ref.sync_loc(
                    es.Identifier(name=first), jac_node=node.parts[0]
                )
                for idx, part in enumerate(parts[1:], start=1):
                    expr = self.pass_ref.sync_loc(
                        es.MemberExpression(
                            object=expr,
                            property=self.pass_ref.sync_loc(
                                es.Identifier(name=part), jac_node=node.parts[idx]
                            ),
                            computed=False,
                            optional=False,
                        ),
                        jac_node=node,
                    )
            elif first and first[0].isupper():
                # Single part starting with uppercase - component identifier
                expr = self.pass_ref.sync_loc(
                    es.Identifier(name=first), jac_node=node.parts[0]
                )
            else:
                # Single part starting with lowercase - HTML element string
                expr = self.pass_ref.sync_loc(es.Literal(value=first), jac_node=node)
        node.gen.es_ast = expr
        return expr

    def spread_attribute(self, node: uni.JsxSpreadAttribute) -> Expression:
        """Process JSX spread attribute."""
        es = self.es
        expr: Expression = (
            cast("Expression", node.expr.gen.es_ast)
            if node.expr and node.expr.gen.es_ast
            else self.pass_ref.sync_loc(
                es.ObjectExpression(properties=[]), jac_node=node
            )
        )
        node.gen.es_ast = expr
        return expr

    def normal_attribute(self, node: uni.JsxNormalAttribute) -> Property:
        """Process JSX normal attribute."""
        es = self.es
        key_expr = self.pass_ref.sync_loc(
            es.Literal(value=node.name.value), jac_node=node.name
        )
        if node.value is None:
            value_expr = self.pass_ref.sync_loc(es.Literal(value=True), jac_node=node)
        elif isinstance(node.value, uni.String):
            value_expr = self.pass_ref.sync_loc(
                es.Literal(value=node.value.lit_value), jac_node=node.value
            )
        else:
            value_expr = (
                node.value.gen.es_ast
                if node.value.gen.es_ast
                else self.pass_ref.sync_loc(es.Literal(value=None), jac_node=node.value)
            )

        prop = self.pass_ref.sync_loc(
            es.Property(
                key=key_expr,
                value=value_expr,
                kind="init",
                method=False,
                shorthand=False,
                computed=False,
            ),
            jac_node=node,
        )
        node.gen.es_ast = prop
        return prop

    def text(self, node: uni.JsxText) -> Expression:
        """Process JSX text node."""
        es = self.es
        raw_value = (
            node.value.value if isinstance(node.value, uni.Token) else node.value
        )
        expr = self.pass_ref.sync_loc(es.Literal(value=str(raw_value)), jac_node=node)
        node.gen.es_ast = expr
        return expr

    def expression(self, node: uni.JsxExpression) -> Expression:
        """Process JSX expression child."""
        es = self.es
        expr: Expression = (
            cast("Expression", node.expr.gen.es_ast)
            if node.expr and node.expr.gen.es_ast
            else self.pass_ref.sync_loc(es.Literal(value=None), jac_node=node.expr)
        )
        node.gen.es_ast = expr
        return expr


class PyJsxProcessor:
    """Generate Python AST structures for JSX nodes."""

    def __init__(self, pass_ref: PyastGenPass) -> None:
        self.pass_ref = pass_ref

    def element(self, node: uni.JsxElement) -> list[ast3.AST]:
        """Generate Python AST for JSX elements."""
        if node.is_fragment or not node.name:
            tag_arg: ast3.expr = self.pass_ref.sync(ast3.Constant(value=None), node)
        else:
            tag_arg = cast(ast3.expr, node.name.gen.py_ast[0])

        if not node.attributes:
            attrs_expr = self.pass_ref.sync(ast3.Dict(keys=[], values=[]), node)
        elif any(isinstance(attr, uni.JsxSpreadAttribute) for attr in node.attributes):
            attrs_expr = self.pass_ref.sync(ast3.Dict(keys=[], values=[]), node)
            for attr in node.attributes:
                if isinstance(attr, uni.JsxSpreadAttribute):
                    attr_ast = cast(ast3.expr, attr.gen.py_ast[0])
                    attrs_expr = self.pass_ref.sync(
                        ast3.Dict(keys=[None, None], values=[attrs_expr, attr_ast]),
                        attr,
                    )
                elif isinstance(attr, uni.JsxNormalAttribute):
                    attr_tuple = cast(ast3.Tuple, attr.gen.py_ast[0])
                    key_ast, value_ast = attr_tuple.elts
                    attrs_expr = self.pass_ref.sync(
                        ast3.Dict(
                            keys=[None, key_ast],
                            values=[attrs_expr, cast(ast3.expr, value_ast)],
                        ),
                        attr,
                    )
        else:
            keys: list[ast3.expr | None] = []
            values: list[ast3.expr] = []
            for attr in node.attributes:
                if isinstance(attr, uni.JsxNormalAttribute):
                    attr_ast = cast(ast3.Tuple, attr.gen.py_ast[0])
                    key_ast, value_ast = attr_ast.elts
                    keys.append(cast(ast3.expr, key_ast))
                    values.append(cast(ast3.expr, value_ast))
            attrs_expr = self.pass_ref.sync(ast3.Dict(keys=keys, values=values), node)

        if node.children:
            children_arg = self.pass_ref.sync(
                ast3.List(
                    elts=[cast(ast3.expr, c.gen.py_ast[0]) for c in node.children],
                    ctx=ast3.Load(),
                ),
                node,
            )
        else:
            children_arg = self.pass_ref.sync(ast3.List(elts=[], ctx=ast3.Load()), node)

        call = self.pass_ref.sync(
            ast3.Call(
                func=self.pass_ref.jaclib_obj("jsx"),
                args=[tag_arg, attrs_expr, children_arg],
                keywords=[],
            ),
            node,
        )
        node.gen.py_ast = [call]
        return node.gen.py_ast

    def element_name(self, node: uni.JsxElementName) -> list[ast3.AST]:
        """Generate Python AST for JSX element names."""
        name_str = ".".join(part.value for part in node.parts)
        expr: ast3.expr
        if node.parts and node.parts[0].value[0].isupper():
            expr = self.pass_ref.sync(
                ast3.Name(id=name_str, ctx=ast3.Load()),
                node,
            )
        else:
            expr = self.pass_ref.sync(ast3.Constant(value=name_str), node)
        node.gen.py_ast = [expr]
        return node.gen.py_ast

    def spread_attribute(self, node: uni.JsxSpreadAttribute) -> list[ast3.AST]:
        """Generate Python AST for JSX spread attributes."""
        node.gen.py_ast = [cast(ast3.expr, node.expr.gen.py_ast[0])]
        return node.gen.py_ast

    def normal_attribute(self, node: uni.JsxNormalAttribute) -> list[ast3.AST]:
        """Generate Python AST for JSX normal attributes."""
        if not node.name:
            node.gen.py_ast = []
            return node.gen.py_ast

        key_ast = self.pass_ref.sync(ast3.Constant(value=node.name.value), node.name)
        value_ast: ast3.expr
        if node.value and node.value.gen.py_ast:
            value_ast = cast(ast3.expr, node.value.gen.py_ast[0])
        else:
            value_ast = self.pass_ref.sync(ast3.Constant(value=True), node)
        node.gen.py_ast = [
            self.pass_ref.sync(
                ast3.Tuple(elts=[key_ast, value_ast], ctx=ast3.Load()),
                node,
            )
        ]
        return node.gen.py_ast

    def text(self, node: uni.JsxText) -> list[ast3.AST]:
        """Generate Python AST for JSX text nodes."""
        raw_value = (
            node.value.value if isinstance(node.value, uni.Token) else node.value
        )
        expr = self.pass_ref.sync(ast3.Constant(value=str(raw_value)), node)
        node.gen.py_ast = [expr]
        return node.gen.py_ast

    def expression(self, node: uni.JsxExpression) -> list[ast3.AST]:
        """Generate Python AST for JSX expression children."""
        node.gen.py_ast = [cast(ast3.expr, node.expr.gen.py_ast[0])]
        return node.gen.py_ast


__all__ = ["EsJsxProcessor", "PyJsxProcessor"]



from oauthlib.oauth2.rfc6749.errors import FatalClientError, OAuth2Error

class FatalOpenIDClientError(FatalClientError): ...
class OpenIDClientError(OAuth2Error): ...

class InteractionRequired(OpenIDClientError):
    error: str
    status_code: int

class LoginRequired(OpenIDClientError):
    error: str
    status_code: int

class AccountSelectionRequired(OpenIDClientError):
    error: str

class ConsentRequired(OpenIDClientError):
    error: str
    status_code: int

class InvalidRequestURI(OpenIDClientError):
    error: str
    description: str

class InvalidRequestObject(OpenIDClientError):
    error: str
    description: str

class RequestNotSupported(OpenIDClientError):
    error: str
    description: str

class RequestURINotSupported(OpenIDClientError):
    error: str
    description: str

class RegistrationNotSupported(OpenIDClientError):
    error: str
    description: str

class InvalidTokenError(OAuth2Error):
    error: str
    status_code: int
    description: str

class InsufficientScopeError(OAuth2Error):
    error: str
    status_code: int
    description: str

def raise_from_error(error: object, params: dict[str, str] | None = None) -> None: ...



from collections.abc import Generator
from typing import overload

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["bridges", "has_bridges", "local_bridges"]

@_dispatchable
def bridges(G: Graph[_Node], root: _Node | None = None) -> Generator[_Node, None, None]: ...
@_dispatchable
def has_bridges(G: Graph[_Node], root: _Node | None = None) -> bool: ...
@overload
def local_bridges(
    G: Graph[_Node], with_span: bool = True, weight: str | None = None
) -> Generator[tuple[_Node, _Node], None, None]: ...
@overload
def local_bridges(
    G: Graph[_Node], with_span: bool = True, weight: str | None = None
) -> Generator[tuple[_Node, _Node, int], None, None]: ...



import logging
from _typeshed import Unused
from argparse import ArgumentParser
from typing import TypedDict, type_check_only

from channels.layers import BaseChannelLayer
from channels.worker import Worker
from django.core.management.base import BaseCommand

logger: logging.Logger

@type_check_only
class _RunWorkerCommandOption(TypedDict):
    verbosity: int | None
    layer: str
    channels: list[str]

class Command(BaseCommand):
    leave_locale_alone: bool = True
    worker_class: type[Worker] = ...
    verbosity: int
    channel_layer: BaseChannelLayer

    def add_arguments(self, parser: ArgumentParser) -> None: ...
    def handle(self, *args: Unused, **options: _RunWorkerCommandOption) -> None: ...



from typing import NamedTuple

from tensorflow.config import experimental as experimental

class PhysicalDevice(NamedTuple):
    name: str
    device_type: str

def list_physical_devices(device_type: None | str = None) -> list[PhysicalDevice]: ...
def get_visible_devices(device_type: None | str = None) -> list[PhysicalDevice]: ...
def set_visible_devices(devices: list[PhysicalDevice] | PhysicalDevice, device_type: None | str = None) -> None: ...
def __getattr__(name: str): ...  # incomplete module



from collections.abc import Callable
from typing import Literal, overload

import numpy as np
from numpy.typing import NDArray

from ._typing import ArrayLikeSeq, GeoArray, GeoT, OptGeoArrayLike, OptGeoArrayLikeSeq, OptGeoT

__all__ = ["transform", "count_coordinates", "get_coordinates", "set_coordinates"]

@overload
def transform(
    geometry: OptGeoT,
    transformation: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    include_z: bool = False,
    *,
    interleaved: bool = True,
) -> OptGeoT: ...
@overload
def transform(
    geometry: OptGeoArrayLikeSeq,
    transformation: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    include_z: bool = False,
    *,
    interleaved: bool = True,
) -> GeoArray: ...
def count_coordinates(geometry: OptGeoArrayLike) -> int: ...
@overload
def get_coordinates(
    geometry: OptGeoArrayLike, include_z: bool = False, return_index: Literal[False] = False, *, include_m: bool = False
) -> NDArray[np.float64]: ...
@overload
def get_coordinates(
    geometry: OptGeoArrayLike, include_z: bool = False, *, return_index: Literal[True], include_m: bool = False
) -> tuple[NDArray[np.float64], NDArray[np.int64]]: ...
@overload
def get_coordinates(
    geometry: OptGeoArrayLike, include_z: bool, return_index: Literal[True], *, include_m: bool = False
) -> tuple[NDArray[np.float64], NDArray[np.int64]]: ...
@overload
def get_coordinates(
    geometry: OptGeoArrayLike, include_z: bool = False, *, return_index: bool, include_m: bool = False
) -> NDArray[np.float64] | tuple[NDArray[np.float64], NDArray[np.int64]]: ...
@overload
def get_coordinates(
    geometry: OptGeoArrayLike, include_z: bool, return_index: bool, *, include_m: bool = False
) -> NDArray[np.float64] | tuple[NDArray[np.float64], NDArray[np.int64]]: ...
@overload
def set_coordinates(geometry: GeoT, coordinates: ArrayLikeSeq[float]) -> GeoT: ...
@overload
def set_coordinates(geometry: OptGeoArrayLikeSeq, coordinates: ArrayLikeSeq[float]) -> GeoArray: ...



from collections.abc import Callable, Container
from typing import TypeVar

_F = TypeVar("_F", bound=Callable[..., object])

class requires_geos:
    version: tuple[int, int, int]
    def __init__(self, version: str) -> None: ...
    def __call__(self, func: _F) -> _F: ...

def multithreading_enabled(func: _F) -> _F: ...
def deprecate_positional(should_be_kwargs: Container[str], category: type[Warning] = ...) -> Callable[..., object]: ...



import os
from _typeshed import Incomplete
from collections.abc import Callable, Generator, Iterable, Mapping
from typing import IO, Any, Literal, TypeVar
from typing_extensions import Concatenate, ParamSpec, Self, TypeAlias, deprecated

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.backend_bases import MouseEvent, RendererBase
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from matplotlib.font_manager import FontProperties
from matplotlib.gridspec import SubplotSpec
from matplotlib.legend import Legend
from matplotlib.patches import Patch
from matplotlib.path import Path as mpl_Path
from matplotlib.patheffects import AbstractPathEffect
from matplotlib.scale import ScaleBase
from matplotlib.text import Text
from matplotlib.transforms import Bbox, BboxBase, Transform, TransformedPath
from matplotlib.typing import ColorType, LineStyleType, MarkerType
from numpy.typing import ArrayLike, NDArray
from pandas import DataFrame, Series

from ._core.typing import ColumnName, DataSource, NormSpec, SupportsDataFrame
from .palettes import _RGBColorPalette
from .utils import _DataSourceWideForm, _Palette, _Vector

__all__ = ["FacetGrid", "PairGrid", "JointGrid", "pairplot", "jointplot"]

_P = ParamSpec("_P")
_R = TypeVar("_R")

_LiteralFont: TypeAlias = Literal["xx-small", "x-small", "small", "medium", "large", "x-large", "xx-large"]

class _BaseGrid:
    def set(
        self,
        *,
        # Keywords follow `matplotlib.axes.Axes.set`. Each keyword <KW> corresponds to a `set_<KW>` method
        adjustable: Literal["box", "datalim"] = ...,
        agg_filter: Callable[[ArrayLike, float], tuple[NDArray[np.floating[Any]], float, float]] | None = ...,
        alpha: float | None = ...,
        anchor: str | tuple[float, float] = ...,
        animated: bool = ...,
        aspect: float | Literal["auto", "equal"] = ...,
        autoscale_on: bool = ...,
        autoscalex_on: bool = ...,
        autoscaley_on: bool = ...,
        axes_locator: Callable[[Axes, RendererBase], Bbox] = ...,
        axisbelow: bool | Literal["line"] = ...,
        box_aspect: float | None = ...,
        clip_box: BboxBase | None = ...,
        clip_on: bool = ...,
        clip_path: Patch | mpl_Path | TransformedPath | None = ...,
        facecolor: ColorType | None = ...,
        frame_on: bool = ...,
        gid: str | None = ...,
        in_layout: bool = ...,
        label: object = ...,
        mouseover: bool = ...,
        navigate: bool = ...,
        path_effects: list[AbstractPathEffect] = ...,
        picker: bool | float | Callable[[Artist, MouseEvent], tuple[bool, dict[Any, Any]]] | None = ...,
        position: Bbox | tuple[float, float, float, float] = ...,
        prop_cycle=...,  # TODO: use cycler.Cycler when cycler gets typed
        rasterization_zorder: float | None = ...,
        rasterized: bool = ...,
        sketch_params: float | None = ...,
        snap: bool | None = ...,
        subplotspec: SubplotSpec = ...,
        title: str = ...,
        transform: Transform | None = ...,
        url: str | None = ...,
        visible: bool = ...,
        xbound: float | None | tuple[float | None, float | None] = ...,
        xlabel: str = ...,
        xlim: float | None | tuple[float | None, float | None] = ...,
        xmargin: float = ...,
        xscale: str | ScaleBase = ...,
        xticklabels: Iterable[str | Text] = ...,
        xticks: ArrayLike = ...,
        ybound: float | None | tuple[float | None, float | None] = ...,
        ylabel: str = ...,
        ylim: float | None | tuple[float | None, float | None] = ...,
        ymargin: float = ...,
        yscale: str | ScaleBase = ...,
        yticklabels: Iterable[str | Text] = ...,
        yticks: ArrayLike = ...,
        zorder: float = ...,
        **kwargs: Any,
    ) -> Self: ...
    @property
    @deprecated("Attribute `fig` is deprecated in favor of `figure`")
    def fig(self) -> Figure: ...
    @property
    def figure(self) -> Figure: ...
    def apply(self, func: Callable[Concatenate[Self, _P], object], *args: _P.args, **kwargs: _P.kwargs) -> Self: ...
    def pipe(self, func: Callable[Concatenate[Self, _P], _R], *args: _P.args, **kwargs: _P.kwargs) -> _R: ...
    def savefig(
        self,
        # Signature follows `matplotlib.figure.Figure.savefig`
        fname: str | os.PathLike[Any] | IO[Any],
        *,
        transparent: bool | None = None,
        dpi: float | Literal["figure"] | None = 96,
        facecolor: ColorType | Literal["auto"] | None = "auto",
        edgecolor: ColorType | Literal["auto"] | None = "auto",
        orientation: Literal["landscape", "portrait"] = "portrait",
        format: str | None = None,
        bbox_inches: Literal["tight"] | Bbox | None = "tight",
        pad_inches: float | Literal["layout"] | None = None,
        backend: str | None = None,
        **kwargs: Any,
    ) -> None: ...

class Grid(_BaseGrid):
    def __init__(self) -> None: ...
    def tight_layout(
        self,
        *,
        # Keywords follow `matplotlib.figure.Figure.tight_layout`
        pad: float = 1.08,
        h_pad: float | None = None,
        w_pad: float | None = None,
        rect: tuple[float, float, float, float] | None = None,
    ) -> Self: ...
    def add_legend(
        self,
        # Cannot use precise key type with union for legend_data because of invariant Mapping keys
        legend_data: Mapping[Any, Artist] | None = None,
        title: str | None = None,
        label_order: list[str] | None = None,
        adjust_subtitles: bool = False,
        *,
        # Keywords follow `matplotlib.legend.Legend`
        loc: str | int | tuple[float, float] | None = None,
        numpoints: int | None = None,
        markerscale: float | None = None,
        markerfirst: bool = True,
        reverse: bool = False,
        scatterpoints: int | None = None,
        scatteryoffsets: Iterable[float] | None = None,
        prop: FontProperties | dict[str, Any] | None = None,
        fontsize: int | _LiteralFont | None = None,
        labelcolor: str | Iterable[str] | None = None,
        borderpad: float | None = None,
        labelspacing: float | None = None,
        handlelength: float | None = None,
        handleheight: float | None = None,
        handletextpad: float | None = None,
        borderaxespad: float | None = None,
        columnspacing: float | None = None,
        ncols: int = 1,
        mode: Literal["expand"] | None = None,
        fancybox: bool | None = None,
        shadow: bool | dict[str, int] | dict[str, float] | None = None,
        title_fontsize: int | _LiteralFont | None = None,
        framealpha: float | None = None,
        edgecolor: ColorType | None = None,
        facecolor: ColorType | None = None,
        bbox_to_anchor: BboxBase | tuple[float, float] | tuple[float, float, float, float] | None = None,
        bbox_transform: Transform | None = None,
        frameon: bool | None = None,
        handler_map: None = None,
        title_fontproperties: FontProperties | None = None,
        alignment: Literal["center", "left", "right"] = "center",
        ncol: int = 1,
        draggable: bool = False,
    ) -> Self: ...
    @property
    def legend(self) -> Legend | None: ...
    def tick_params(
        self,
        axis: Literal["x", "y", "both"] = "both",
        *,
        # Keywords follow `matplotlib.axes.Axes.tick_params`
        which: Literal["major", "minor", "both"] = "major",
        reset: bool = False,
        direction: Literal["in", "out", "inout"] = ...,
        length: float = ...,
        width: float = ...,
        color: ColorType = ...,
        pad: float = ...,
        labelsize: float | str = ...,
        labelcolor: ColorType = ...,
        labelfontfamily: str = ...,
        colors: ColorType = ...,
        zorder: float = ...,
        bottom: bool = ...,
        top: bool = ...,
        left: bool = ...,
        right: bool = ...,
        labelbottom: bool = ...,
        labeltop: bool = ...,
        labelleft: bool = ...,
        labelright: bool = ...,
        labelrotation: float = ...,
        grid_color: ColorType = ...,
        grid_alpha: float = ...,
        grid_linewidth: float = ...,
        grid_linestyle: str = ...,
        **kwargs: Any,
    ) -> Self: ...

class FacetGrid(Grid):
    data: DataFrame
    row_names: list[Any]
    col_names: list[Any]
    hue_names: list[Any] | None
    hue_kws: dict[str, Any]
    def __init__(
        self,
        data: DataFrame | SupportsDataFrame,
        *,
        row: str | None = None,
        col: str | None = None,
        hue: str | None = None,
        col_wrap: int | None = None,
        sharex: bool | Literal["col", "row"] = True,
        sharey: bool | Literal["col", "row"] = True,
        height: float = 3,
        aspect: float = 1,
        palette: _Palette | None = None,
        row_order: Iterable[Any] | None = None,
        col_order: Iterable[Any] | None = None,
        hue_order: Iterable[Any] | None = None,
        hue_kws: dict[str, Any] | None = None,
        dropna: bool = False,
        legend_out: bool = True,
        despine: bool = True,
        margin_titles: bool = False,
        xlim: tuple[float, float] | None = None,
        ylim: tuple[float, float] | None = None,
        subplot_kws: dict[str, Any] | None = None,
        gridspec_kws: dict[str, Any] | None = None,
    ) -> None: ...
    def facet_data(self) -> Generator[tuple[tuple[int, int, int], DataFrame], None, None]: ...
    def map(self, func: Callable[..., object], *args: str, **kwargs: Any) -> Self: ...
    def map_dataframe(self, func: Callable[..., object], *args: str, **kwargs: Any) -> Self: ...
    def facet_axis(self, row_i: int, col_j: int, modify_state: bool = True) -> Axes: ...
    # `despine` should be kept roughly in line with `seaborn.utils.despine`
    def despine(
        self,
        *,
        ax: Axes | None = None,
        top: bool = True,
        right: bool = True,
        left: bool = False,
        bottom: bool = False,
        offset: int | Mapping[str, int] | None = None,
        trim: bool = False,
    ) -> Self: ...
    def set_axis_labels(
        self, x_var: str | None = None, y_var: str | None = None, clear_inner: bool = True, **kwargs: Any
    ) -> Self: ...
    def set_xlabels(self, label: str | None = None, clear_inner: bool = True, **kwargs: Any) -> Self: ...
    def set_ylabels(self, label: str | None = None, clear_inner: bool = True, **kwargs: Any) -> Self: ...
    def set_xticklabels(self, labels: Iterable[str | Text] | None = None, step: int | None = None, **kwargs: Any) -> Self: ...
    def set_yticklabels(self, labels: Iterable[str | Text] | None = None, **kwargs: Any) -> Self: ...
    def set_titles(
        self, template: str | None = None, row_template: str | None = None, col_template: str | None = None, **kwargs: Any
    ) -> Self: ...
    def refline(
        self,
        *,
        x: float | None = None,
        y: float | None = None,
        color: ColorType = ".5",
        linestyle: LineStyleType = "--",
        **line_kws: Any,
    ) -> Self: ...
    @property
    def axes(self) -> NDArray[Incomplete]: ...  # array of `Axes`
    @property
    def ax(self) -> Axes: ...
    @property
    def axes_dict(self) -> dict[Any, Axes]: ...

class PairGrid(Grid):
    x_vars: list[str]
    y_vars: list[str]
    square_grid: bool
    axes: NDArray[Incomplete]  # two-dimensional array of `Axes`
    data: DataFrame
    diag_sharey: bool
    diag_vars: list[str] | None
    diag_axes: list[Axes] | None
    hue_names: list[str]
    hue_vals: Series[Any]
    hue_kws: dict[str, Any]
    palette: _RGBColorPalette
    def __init__(
        self,
        data: DataFrame | SupportsDataFrame,
        *,
        hue: str | None = None,
        vars: Iterable[str] | None = None,
        x_vars: Iterable[str] | str | None = None,
        y_vars: Iterable[str] | str | None = None,
        hue_order: Iterable[str] | None = None,
        palette: _Palette | None = None,
        hue_kws: dict[str, Any] | None = None,
        corner: bool = False,
        diag_sharey: bool = True,
        height: float = 2.5,
        aspect: float = 1,
        layout_pad: float = 0.5,
        despine: bool = True,
        dropna: bool = False,
    ) -> None: ...
    def map(self, func: Callable[..., object], **kwargs: Any) -> Self: ...
    def map_lower(self, func: Callable[..., object], **kwargs: Any) -> Self: ...
    def map_upper(self, func: Callable[..., object], **kwargs: Any) -> Self: ...
    def map_offdiag(self, func: Callable[..., object], **kwargs: Any) -> Self: ...
    def map_diag(self, func: Callable[..., object], **kwargs: Any) -> Self: ...

class JointGrid(_BaseGrid):
    ax_joint: Axes
    ax_marg_x: Axes
    ax_marg_y: Axes
    x: Series[Any]
    y: Series[Any]
    hue: Series[Any]
    def __init__(
        self,
        data: DataSource | _DataSourceWideForm | None = None,
        *,
        x: ColumnName | _Vector | None = None,
        y: ColumnName | _Vector | None = None,
        hue: ColumnName | _Vector | None = None,
        height: float = 6,
        ratio: float = 5,
        space: float = 0.2,
        palette: _Palette | Colormap | None = None,
        hue_order: Iterable[ColumnName] | None = None,
        hue_norm: NormSpec = None,
        dropna: bool = False,
        xlim: float | tuple[float, float] | None = None,
        ylim: float | tuple[float, float] | None = None,
        marginal_ticks: bool = False,
    ) -> None: ...
    def plot(self, joint_func: Callable[..., object], marginal_func: Callable[..., object], **kwargs: Any) -> Self: ...
    def plot_joint(self, func: Callable[..., object], **kwargs: Any) -> Self: ...
    def plot_marginals(self, func: Callable[..., object], **kwargs: Any) -> Self: ...
    def refline(
        self,
        *,
        x: float | None = None,
        y: float | None = None,
        joint: bool = True,
        marginal: bool = True,
        color: ColorType = ".5",
        linestyle: LineStyleType = "--",
        **line_kws: Any,
    ) -> Self: ...
    def set_axis_labels(self, xlabel: str = "", ylabel: str = "", **kwargs: Any) -> Self: ...

def pairplot(
    data: DataFrame,
    *,
    hue: str | None = None,
    hue_order: Iterable[str] | None = None,
    palette: _Palette | None = None,
    vars: Iterable[str] | None = None,
    x_vars: Iterable[str] | str | None = None,
    y_vars: Iterable[str] | str | None = None,
    kind: Literal["scatter", "kde", "hist", "reg"] = "scatter",
    diag_kind: Literal["auto", "hist", "kde"] | None = "auto",
    markers: MarkerType | list[MarkerType] | None = None,
    height: float = 2.5,
    aspect: float = 1,
    corner: bool = False,
    dropna: bool = False,
    plot_kws: dict[str, Any] | None = None,
    diag_kws: dict[str, Any] | None = None,
    grid_kws: dict[str, Any] | None = None,
    size: float | None = None,  # deprecated
) -> PairGrid: ...
def jointplot(
    data: DataSource | _DataSourceWideForm | None = None,
    *,
    x: ColumnName | _Vector | None = None,
    y: ColumnName | _Vector | None = None,
    hue: ColumnName | _Vector | None = None,
    kind: Literal["scatter", "kde", "hist", "hex", "reg", "resid"] = "scatter",
    height: float = 6,
    ratio: float = 5,
    space: float = 0.2,
    dropna: bool = False,
    xlim: float | tuple[float, float] | None = None,
    ylim: float | tuple[float, float] | None = None,
    color: ColorType | None = None,
    palette: _Palette | Colormap | None = None,
    hue_order: Iterable[ColumnName] | None = None,
    hue_norm: NormSpec = None,
    marginal_ticks: bool = False,
    joint_kws: dict[str, Any] | None = None,
    marginal_kws: dict[str, Any] | None = None,
    **kwargs: Any,
) -> JointGrid: ...



from _typeshed.wsgi import WSGIEnvironment
from collections.abc import Iterator, MutableMapping
from typing import TypeVar, overload

from webob.multidict import MultiDict

__all__ = ["ResponseHeaders", "EnvironHeaders"]

_T = TypeVar("_T")

class ResponseHeaders(MultiDict[str, str]):
    def __getitem__(self, key: str) -> str: ...
    def getall(self, key: str) -> list[str]: ...
    def mixed(self) -> dict[str, str | list[str]]: ...
    def dict_of_lists(self) -> dict[str, list[str]]: ...
    def __setitem__(self, key: str, value: str) -> None: ...
    def __delitem__(self, key: str) -> None: ...
    def __contains__(self, key: object) -> bool: ...
    has_key = __contains__
    def setdefault(self, key: str, default: str) -> str: ...  # type: ignore[override]
    @overload
    def pop(self, key: str) -> str: ...
    @overload
    def pop(self, key: str, default: _T) -> str | _T: ...

class EnvironHeaders(MutableMapping[str, str]):
    environ: WSGIEnvironment
    def __init__(self, environ: WSGIEnvironment) -> None: ...
    def __getitem__(self, hname: str) -> str: ...
    def __setitem__(self, hname: str, value: str) -> None: ...
    def __delitem__(self, hname: str) -> None: ...
    def keys(self) -> Iterator[str]: ...  # type: ignore[override]
    def __contains__(self, hname: object) -> bool: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[str]: ...
